import copy

import numpy as np
import tensorflow as tf
import tensorflow.keras as tfk
import tensorflow.keras.layers as tfkl
from tensorflow.python.eager import context
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import tensor_shape
from tensorflow.python.keras import constraints
from tensorflow.python.keras import initializers
from tensorflow.python.keras.engine.input_spec import InputSpec
from tensorflow.python.ops import gen_math_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import nn
from tensorflow.python.ops import sparse_ops
from tensorflow.python.ops import standard_ops

from dynastes import activations
from dynastes import regularizers
from dynastes import weight_normalizers


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class _WscaleInitializer(tfk.initializers.Initializer):

    def __init__(self,
                 initializer,
                 lrmul=1.,
                 **kwargs):
        super(_WscaleInitializer, self).__init__(**kwargs)
        self.initializer = tfk.initializers.get(initializer)
        self.lrmul = lrmul

    def __call__(self, shape, dtype=None):
        return self.initializer(shape, dtype=dtype) / self.lrmul

    def get_config(self):
        config = {
            'initializer': tfk.initializers.serialize(self.initializer),
            'lrmul': self.lrmul,
        }
        base_config = super(_WscaleInitializer, self).get_config()
        return {**base_config, **config}


def _get_regularizers_from_keywords(kwargs):
    _initializers = {}
    _regularizers = {}
    _constraints = {}
    _normalizers = {}

    kwarg_keys = copy.copy(list(kwargs.keys()))

    for kwarg in kwarg_keys:
        if kwarg.endswith('initializer'):
            _initializers[kwarg.split('_initializer')[0]] = initializers.get(kwargs.pop(kwarg, None))
        elif kwarg.endswith('regularizer'):
            if kwarg != 'activity_regularizer':
                _regularizers[kwarg.split('_regularizer')[0]] = regularizers.get(kwargs.pop(kwarg, None))
        elif kwarg.endswith('constraint'):
            _constraints[kwarg.split('_constraint')[0]] = constraints.get(kwargs.pop(kwarg, None))
        elif kwarg.endswith('normalizer'):
            _normalizers[kwarg.split('_normalizer')[0]] = weight_normalizers.get(kwargs.pop(kwarg, None))

    return _initializers, _regularizers, _constraints, _normalizers


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class DynastesBaseLayer(tfkl.Layer):
    def __init__(self,
                 activity_regularizer=None,
                 trainable=True,
                 name=None,
                 use_wscale=False,
                 wlrmul=1.,
                 wgain=np.sqrt(2),
                 supports_caching=False,
                 **kwargs):
        self.weights_dict = {}
        self.initializers, self.regularizers, self.constraints, self.normalizers = _get_regularizers_from_keywords(
            kwargs)
        self.use_wscale = use_wscale
        self.lrmul = wlrmul
        self.gain = wgain
        self.supports_caching = supports_caching
        super(DynastesBaseLayer, self).__init__(
            trainable=trainable,
            name=name,
            activity_regularizer=regularizers.get(activity_regularizer),
            **kwargs)
        self.supports_masking = True

    def request_cache(self, batch_size, **kwargs):
        pass

    def get_initializer(self, name):
        if name not in self.initializers:
            if name == 'kernel':
                self.initializers['kernel'] = initializers.get('he_uniform')
            elif name == 'bias':
                self.initializers['bias'] = initializers.get('zeros')
            else:
                self.initializers[name] = initializers.get('glorot_uniform')

        return self.initializers[name]

    def get_regularizer(self, name):
        if name not in self.regularizers:
            self.regularizers[name] = regularizers.get(None)
        return self.regularizers[name]

    def get_constraint(self, name):
        if name not in self.constraints:
            self.constraints[name] = constraints.get(None)
        return self.constraints[name]

    def add_weight(self,
                   name=None,
                   shape=None,
                   trainable=None,
                   partitioner=None,
                   initializer=None,
                   regularizer=None,
                   constraint=None,
                   dtype=None,
                   use_resource=None,
                   **kwargs):
        if initializer is not None:
            self.initializers[name] = initializers.get(initializer)
        if regularizer is not None:
            self.regularizers[name] = regularizers.get(regularizer)
        if constraint is not None:
            self.constraints[name] = constraints.get(constraint)
        _initializer = self.get_initializer(name)
        if self.use_wscale:
            _initializer = _WscaleInitializer(_initializer, lrmul=self.lrmul)
            self.initializers[name] = _initializer
            if name in self.normalizers and self.normalizers[name] is not None:
                self.normalizers[name] = weight_normalizers.WscaleNormalizer(next_layer=self.normalizers[name],
                                                                             lrmul=self.lrmul,
                                                                             gain=self.gain)
            else:
                self.normalizers[name] = weight_normalizers.WscaleNormalizer(lrmul=self.lrmul, gain=self.gain)

        weight = super(DynastesBaseLayer, self).add_weight(name=name,
                                                           shape=shape,
                                                           initializer=_initializer,
                                                           regularizer=self.get_regularizer(name),
                                                           trainable=trainable,
                                                           constraint=self.get_constraint(name),
                                                           partitioner=partitioner,
                                                           use_resource=use_resource,
                                                           **kwargs)
        if name in self.normalizers:
            if self.normalizers[name] is not None:
                self.normalizers[name].build(shape)
        self.weights_dict[name] = weight
        return weight

    def get_weight(self, name, training=None):
        w = self.weights_dict[name]
        if name in self.normalizers and self.normalizers[name] is not None:
            w = self.normalizers[name](w, training=training)
        return w

    def get_config(self):
        config = {
            'activity_regularizer': regularizers.serialize(self.activity_regularizer),
            'use_wscale': self.use_wscale,
            'wlrmul': self.lrmul,
            'wgain': self.gain,
        }
        for name, initializer in self.initializers.items():
            config[name + '_initializer'] = initializers.serialize(initializer)
        for name, regularizer in self.regularizers.items():
            config[name + '_regularizer'] = regularizers.serialize(regularizer)
        for name, constraint in self.constraints.items():
            config[name + '_constraint'] = constraints.serialize(constraint)
        for name, normalizer in self.normalizers.items():
            config[name + '_normalizer'] = weight_normalizers.serialize(normalizer)

        base_config = super(DynastesBaseLayer, self).get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class ActivatedKernelBiasBaseLayer(DynastesBaseLayer):
    def __init__(self,
                 activation=None,
                 use_bias=True,
                 activity_regularizer=None,
                 **kwargs):
        super(ActivatedKernelBiasBaseLayer, self).__init__(
            activity_regularizer=regularizers.get(activity_regularizer),
            **kwargs)
        self.activation = activations.get(activation)
        self.use_bias = use_bias

    def build_kernel(self, shape):
        return self.add_weight(
            name='kernel',
            shape=shape,
            trainable=True,
            dtype=self.dtype)

    def build_bias(self, output_dim):
        if self.use_bias:
            return self.add_weight(
                name='bias',
                shape=(output_dim,),
                trainable=True,
                dtype=self.dtype)
        return None

    def post_process_call(self, x, training=None):
        if self.use_bias:
            x = nn.bias_add(x, self.get_weight('bias', training=training), data_format='NHWC')
        if self.activation is not None:
            return self.activation(x)
        return x

    def call(self, x, training=None):
        return self.post_process_call(x, training=training)

    def get_config(self):
        config = {
            'activation': activations.serialize(self.activation),
            'use_bias': self.use_bias,
        }
        base_config = super(ActivatedKernelBiasBaseLayer, self).get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class DynastesDense(ActivatedKernelBiasBaseLayer):
    """Just your regular densely-connected NN layer.
    `Dense` implements the operation:
    `output = activation(dot(input, kernel) + bias)`
    where `activation` is the element-wise activation function
    passed as the `activation` argument, `kernel` is a weights matrix
    created by the layer, and `bias` is a bias vector created by the layer
    (only applicable if `use_bias` is `True`).
    Note: If the input to the layer has a rank greater than 2, then
    it is flattened prior to the initial dot product with `kernel`.
    Example:
    ```python
    # as first layer in a sequential model:
    model = Sequential()
    model.add(Dense(32, input_shape=(16,)))
    # now the model will take as input arrays of shape (*, 16)
    # and output arrays of shape (*, 32)
    # after the first layer, you don't need to specify
    # the size of the input anymore:
    model.add(Dense(32))
    ```
    Arguments:
      units: Positive integer, dimensionality of the output space.
      activation: Activation function to use.
        If you don't specify anything, no activation is applied
        (ie. "linear" activation: `a(x) = x`).
      use_bias: Boolean, whether the layer uses a bias vector.
      kernel_initializer: Initializer for the `kernel` weights matrix.
      bias_initializer: Initializer for the bias vector.
      kernel_regularizer: Regularizer function applied to
        the `kernel` weights matrix.
      bias_regularizer: Regularizer function applied to the bias vector.
      activity_regularizer: Regularizer function applied to
        the output of the layer (its "activation")..
      kernel_constraint: Constraint function applied to
        the `kernel` weights matrix.
      bias_constraint: Constraint function applied to the bias vector.
    Input shape:
      N-D tensor with shape: `(batch_size, ..., input_dim)`.
      The most common situation would be
      a 2D input with shape `(batch_size, input_dim)`.
    Output shape:
      N-D tensor with shape: `(batch_size, ..., units)`.
      For instance, for a 2D input with shape `(batch_size, input_dim)`,
      the output would have shape `(batch_size, units)`.
    """

    def __init__(self,
                 units,
                 **kwargs):
        if 'input_shape' not in kwargs and 'input_dim' in kwargs:
            kwargs['input_shape'] = (kwargs.pop('input_dim'),)

        super(DynastesDense, self).__init__(**kwargs)
        self.units = int(units)

        self.supports_masking = True
        self.input_spec = InputSpec(min_ndim=2)

    def build(self, input_shape):
        dtype = dtypes.as_dtype(self.dtype or tfk.backend.floatx())
        if not (dtype.is_floating or dtype.is_complex):
            raise TypeError('Unable to build `Dense` layer with non-floating point '
                            'dtype %s' % (dtype,))
        input_shape = tensor_shape.TensorShape(input_shape)
        if tensor_shape.dimension_value(input_shape[-1]) is None:
            raise ValueError('The last dimension of the inputs to `Dense` '
                             'should be defined. Found `None`.')
        last_dim = tensor_shape.dimension_value(input_shape[-1])
        self.input_spec = InputSpec(min_ndim=2,
                                    axes={-1: last_dim})
        self.kernel = self.build_kernel(
            shape=[last_dim, self.units])
        self.build_bias(self.units)
        self.built = True

    def call(self, inputs, training=None):
        rank = len(inputs.shape)
        kernel = self.get_weight('kernel', training=training)
        if rank > 2:
            # Broadcasting is required for the inputs.
            outputs = standard_ops.tensordot(inputs, kernel, [[rank - 1], [0]])
            # Reshape the output back to the original ndim of the input.
            if not context.executing_eagerly():
                shape = inputs.shape.as_list()
                output_shape = shape[:-1] + [self.units]
                outputs.set_shape(output_shape)
        else:
            inputs = math_ops.cast(inputs, self._compute_dtype)
            if tfk.backend.is_sparse(inputs):
                outputs = sparse_ops.sparse_tensor_dense_matmul(inputs, kernel)
            else:
                outputs = gen_math_ops.mat_mul(inputs, kernel)
        return super(DynastesDense, self).call(outputs, training=training)

    def compute_output_shape(self, input_shape):
        input_shape = tensor_shape.TensorShape(input_shape)
        input_shape = input_shape.with_rank_at_least(2)
        if tensor_shape.dimension_value(input_shape[-1]) is None:
            raise ValueError(
                'The innermost dimension of input_shape must be defined, but saw: %s'
                % input_shape)
        return input_shape[:-1].concatenate(self.units)

    def get_config(self):
        config = {
            'units': self.units,
        }
        base_config = super(DynastesDense, self).get_config()
        return {**base_config, **config}



import tensorflow as tf
from tensorflow.keras.utils import custom_object_scope

from dynastes.blocks.attention_blocks import SelfAttentionBlock1D
from dynastes.util.test_utils import layer_test


class SelfAttentionBlock1DTest(tf.test.TestCase):
    def test_simple(self):
            layer_test(
                SelfAttentionBlock1D, kwargs={'attention_dim': 8, 'kernel_size': 3, 'output_dim': 16, 'num_heads': 4,
                                              'multiquery_attention': True}, input_shape=(5, 32, 3))



import tensorflow as tf
from tensorflow.python.framework import ops
from tensorflow.python.util.compat import collections_abc


def _get_sequence(value, n, channel_index, name):
    """Formats a value input for gen_nn_ops."""
    if value is None:
        value = [1]
    elif not isinstance(value, collections_abc.Sized):
        value = [value]

    current_n = len(value)
    if current_n == n + 2:
        return value
    elif current_n == 1:
        value = list((value[0],) * n)
    elif current_n == n:
        value = list(value)
    else:
        raise ValueError("{} should be of length 1, {} or {} but was {}".format(
            name, n, n + 2, current_n))

    if channel_index == 1:
        return [1, 1] + value
    else:
        return [1] + value + [1]


def depthwise_conv2d_transpose(
        input,  # pylint: disable=redefined-builtin
        filters,  # pylint: disable=redefined-builtin
        output_shape,
        strides,
        padding="SAME",
        data_format="NHWC",
        dilations=None,
        name=None):
    """The transpose of `depthwise_conv2d`.
    This operation is sometimes called "deconvolution" after [Deconvolutional
    Networks](http://www.matthewzeiler.com/pubs/cvpr2010/cvpr2010.pdf), but is
    actually the transpose (gradient) of `depthwise_conv2d` rather than an actual
    deconvolution.
    Args:
      input: A 4-D `Tensor` of type `float` and shape `[batch, height, width,
        in_channels]` for `NHWC` data format or `[batch, in_channels, height,
        width]` for `NCHW` data format.
      filters: A 4-D `Tensor` with the same type as `input` and shape `[height,
        width, channel_divider, in_channels]`.  `filter`'s `in_channels` dimension
        must match that of `input`.
      output_shape: A 1-D `Tensor` representing the output shape of the
        deconvolution op.
      strides: An int or list of `ints` that has length `1`, `2` or `4`.  The
        stride of the sliding window for each dimension of `input`. If a single
        value is given it is replicated in the `H` and `W` dimension. By default
        the `N` and `C` dimensions are set to 0. The dimension order is determined
        by the value of `data_format`, see below for details.
      padding: A string, either `'VALID'` or `'SAME'`. The padding algorithm. See
        the "returns" section of `tf.nn.convolution` for details.
      data_format: A string. 'NHWC' and 'NCHW' are supported.
      dilations: An int or list of `ints` that has length `1`, `2` or `4`,
        defaults to 1. The dilation factor for each dimension of`input`. If a
        single value is given it is replicated in the `H` and `W` dimension. By
        default the `N` and `C` dimensions are set to 1. If set to k > 1, there
        will be k-1 skipped cells between each filter element on that dimension.
        The dimension order is determined by the value of `data_format`, see above
        for details. Dilations in the batch and depth dimensions if a 4-d tensor
        must be 1.
      name: Optional name for the returned tensor.
    Returns:
      A `Tensor` with the same type as `input`.
    Raises:
      ValueError: If input/output depth does not match `filter`'s shape, or if
        padding is other than `'VALID'` or `'SAME'`.
    """
    with ops.name_scope(name, "depthwise_conv2d_transpose",
                        [input, filter, output_shape]) as name:
        if data_format is None:
            data_format = "NHWC"
        channel_index = 1 if data_format.startswith("NC") else 3

        strides = _get_sequence(strides, 2, channel_index, "strides")
        dilations = _get_sequence(dilations, 2, channel_index, "dilations")

        return tf.nn.depthwise_conv2d_backprop_input(
            input_sizes=output_shape,
            filter=filters,
            out_backprop=input,
            strides=strides,
            padding=padding,
            data_format=data_format,
            dilations=dilations,
            name=name)



import numpy as np
import tensorflow as tf
from tensorflow.python.framework import test_util
from tensorflow_addons.layers.normalizations import GroupNormalization

from dynastes.layers.normalization_layers import AdaptiveMultiNormalization, PoolNormalization2D


def _test_grads(testCase: tf.test.TestCase, func, input):
    _, grads = tf.test.compute_gradient(func, input)
    for grad in grads:
        testCase.assertNotAllClose(grad, np.zeros_like(grad))


to_tensor = tf.convert_to_tensor
normal = np.random.normal


class AdaptiveMultiNormalizationTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        normalizers = [
            GroupNormalization(groups=1, center=False, scale=False),
            GroupNormalization(groups=-1, center=False, scale=False),
            PoolNormalization2D(pool_size=(-1, 3))
        ]
        layer = AdaptiveMultiNormalization(layers=normalizers)
        x = tf.convert_to_tensor(normal(size=(1, 8, 8, 8)).astype(np.float16))
        y = tf.convert_to_tensor(normal(size=(1, 2, 3, 4)).astype(np.float16))
        res = layer([x, y])
        self.assertShapeEqual(x.numpy(), res)
        y = tf.convert_to_tensor(normal(size=(1, 4)).astype(np.float16))
        res = layer([x, y])
        self.assertShapeEqual(x.numpy(), res)



import numpy as np
import tensorflow as tf
from tensorflow.python.framework import test_util
from tensorflow.python.keras.testing_utils import layer_test

from dynastes.layers.random_layers import StatelessRandomNormalLike


def _test_grads(testCase: tf.test.TestCase, func, input):
    _, grads = tf.test.compute_gradient(func, input)
    for grad in grads:
        testCase.assertNotAllClose(grad, np.zeros_like(grad))


to_tensor = tf.convert_to_tensor
normal = np.random.normal


class StatelessRandomNormalLikeTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        layer_test(
            StatelessRandomNormalLike, input_shape=(5, 32, 3))

        layer = StatelessRandomNormalLike(channels=1)
        input = tf.random.normal(shape=[4, 16, 32, 5])
        seed = tf.convert_to_tensor(np.random.randint(tf.int64.min, tf.int64.max, size=(4, 4), dtype=np.int64))
        res = layer([input, seed])
        self.assertNotAllClose(input, res)
        seed = tf.convert_to_tensor([1, 2, 1, 3])
        res = layer([input, seed])
        self.assertAllEqual(res[0], res[2])



import tensorflow as tf
import tensorflow.keras.layers as tfkl

from dynastes.layers.base_layers import DynastesBaseLayer


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class MultiHeadAttentionLayer(DynastesBaseLayer):

    def __init__(self,
                 q_layer: tfkl.Layer,
                 k_layer: tfkl.Layer,
                 v_layer: tfkl.Layer,
                 attention_layer: tfkl.Layer, **kwargs):
        super(MultiHeadAttentionLayer, self).__init__(**kwargs)
        self.q_layer = q_layer
        self.k_layer = k_layer
        self.v_layer = v_layer
        self.attention_layer = attention_layer

    def call(self, inputs, training=None, mask=None):

        if type(inputs) == list:
            if len(inputs) == 3:
                q, k, v = inputs
            else:
                raise ValueError('Incorrect inputs')
        else:
            raise ValueError('Incorrect inputs')

        if mask is not None:
            assert type(mask) == type(inputs)
            if type(mask) == list:
                assert len(mask) == 3
            elif type(mask) == dict:
                mask = [mask['q'], mask['k'], mask['v']]
            else:
                assert False

        q = self.q_layer(q, training=training)
        k = self.k_layer(k, training=training)
        v = self.v_layer(v, training=training)

        return self.attention_layer([q, k, v], mask=mask, training=training)

    def compute_output_shape(self, input_shape):
        assert type(input_shape) == list
        assert len(input_shape) == 3
        qs, ks, vs = input_shape
        qs = self.q_layer.compute_output_shape(qs)
        ks = self.k_layer.compute_output_shape(ks)
        vs = self.v_layer.compute_output_shape(vs)
        return self.attention_layer.compute_output_shape([qs, ks, vs])


class MultiHeadSelfAttentionLayer(DynastesBaseLayer):

    def __init__(self,
                 q_layer,
                 k_layer,
                 v_layer,
                 attention_layer, **kwargs):
        super(MultiHeadSelfAttentionLayer, self).__init__(**kwargs)
        self.multiheadAttentionLayer = MultiHeadAttentionLayer(q_layer=q_layer,
                                                               k_layer=k_layer,
                                                               v_layer=v_layer,
                                                               attention_layer=attention_layer,
                                                               name='MultiHeadAttention')

    def call(self, inputs, mask=None, training=None):
        assert type(inputs) != list
        inputs = [inputs, inputs, inputs]
        if mask is not None:
            assert type(mask) != list
            mask = [mask, mask, mask]
        return self.multiheadAttentionLayer(inputs=inputs, mask=mask, training=training)

    def compute_output_shape(self, input_shape):
        return self.multiheadAttentionLayer.compute_output_shape([input_shape, input_shape, input_shape])


class MultiHeadCrossAttentionLayer(DynastesBaseLayer):

    def __init__(self,
                 q_layer,
                 k_layer,
                 v_layer,
                 attention_layer, **kwargs):
        super(MultiHeadCrossAttentionLayer, self).__init__(**kwargs)
        self.multiheadAttentionLayer = MultiHeadAttentionLayer(q_layer=q_layer,
                                                               k_layer=k_layer,
                                                               v_layer=v_layer,
                                                               attention_layer=attention_layer,
                                                               name='MultiHeadAttention')

    def call(self, inputs, mask=None, training=None):
        assert len(inputs) == 2
        inputs = [inputs[0], inputs[1], inputs[1]]
        if mask is not None:
            mask = [mask[0], mask[1], mask[1]]
        return self.multiheadAttentionLayer(inputs=inputs, mask=mask, training=training)

    def compute_output_shape(self, input_shape):
        return self.multiheadAttentionLayer.compute_output_shape([input_shape[0], input_shape[1], input_shape[1]])



import tensorflow as tf

from dynastes.ops.pad_ops import pad_input_2d
from dynastes.ops.t2t_common import shape_list


def time_delay_nn_1d(x, kernel, kernel_size, strides, dilation_rate, padding='valid'):
    shape = shape_list(x)
    x = tf.expand_dims(x, -1)
    x, padding = pad_input_2d(x, padding, kernel_size=(kernel_size, shape[-1]), dilation_rate=(dilation_rate, 1))
    x = tf.image.extract_patches(x,
                                 sizes=[1, kernel_size, shape[-1], 1],
                                 strides=[1, strides, shape[-1], 1],
                                 rates=[1, dilation_rate, 1, 1],
                                 padding=padding)
    x = tf.squeeze(x, -2)
    x = tf.matmul(x, kernel)
    return x



import numpy as np
import tensorflow as tf
import tensorflow.keras.layers as tfkl
from tensorflow.python.framework import test_util

from dynastes.blocks.attention_blocks import SelfAttentionBlock1D, AttentionBlock1D
from dynastes.blocks.transformer_blocks import PointWiseFeedForwardBlock, EncoderBlock, DecoderBlock, EncoderBlockStack, \
    DecoderBlockStack
from dynastes.util import cache_context


def _test_grads(testCase: tf.test.TestCase, func, input):
    _, grads = tf.test.compute_gradient(func, input)
    for grad in grads:
        testCase.assertNotAllClose(grad, np.zeros_like(grad))
        testCase.assertAllInRange(grad, -400., 400)


to_tensor = tf.convert_to_tensor
normal = np.random.normal


class DecoderBlockTest(tf.test.TestCase):

    @test_util.use_deterministic_cudnn
    def test_simple(self):
        with cache_context.CacheContext():
            d_model = 16
            num_heads = 4
            dff = 32
            max_length = 64
            batch_size = 23
            mask_len = 7
            enc_sablock = SelfAttentionBlock1D(d_model // 4, d_model, num_heads=num_heads,
                                               attention_type='Attention1D',
                                               relative=True,
                                               masked=True,
                                               mask_right=False,
                                               multiquery_attention=True, )
            enc_norm = tfkl.LayerNormalization(epsilon=1e-6)
            enc_df_net = PointWiseFeedForwardBlock(dff=dff, d_model=d_model)
            enc_block = EncoderBlock(sa_layer=enc_sablock, norm0=enc_norm, ffn=enc_df_net, norm1=enc_norm)

            stack = EncoderBlockStack([enc_block] * 3)

            enc_input = tf.convert_to_tensor(normal(size=(batch_size, 32, d_model)).astype(np.float32)).numpy()
            enc_mask = to_tensor(([True] * (32 - mask_len)) + ([False] * (mask_len)))
            enc_mask = tf.expand_dims(enc_mask, axis=0)
            enc_mask = tf.tile(enc_mask, [batch_size, 1])
            enc_out = enc_block(enc_input, training=None, mask=enc_mask)
            encoded_out = stack(enc_input, training=None, mask=enc_mask)
            comp_out_shape = stack.compute_output_shape(enc_input.shape)

            dec_sablock = SelfAttentionBlock1D(d_model // 4, d_model, num_heads=num_heads,
                                               attention_type='Attention1D',
                                               relative=False,
                                               local=True,
                                               masked=True,
                                               mask_right=True,
                                               multiquery_attention=True,
                                               pad_q_to_kv=True)
            dec_cablock = AttentionBlock1D(d_model // 4, d_model, num_heads=num_heads,
                                           attention_type='Attention1D',
                                           relative=False,
                                           masked=True,
                                           mask_right=False,
                                           multiquery_attention=True,
                                           cache_kv=True)
            dec_norm = tfkl.LayerNormalization(epsilon=1e-6)
            dec_df_net = PointWiseFeedForwardBlock(dff=dff, d_model=d_model)
            dec_block = DecoderBlock(sa_layer=dec_sablock, ca_layer=dec_cablock, norm0=dec_norm, ffn=dec_df_net,
                                     norm1=dec_norm, norm2=dec_norm)

            stack = DecoderBlockStack([dec_block] * 3)
            cache = stack.request_cache(batch_size=batch_size, max_length_sa=max_length, max_length_ca=32)

            dec_input = tf.convert_to_tensor(normal(size=(batch_size, 1, d_model)).astype(np.float32)).numpy()

            self.assertEqual(encoded_out.shape, comp_out_shape)
            self.assertEqual(enc_out.shape, comp_out_shape)

            def inc_encode(start, _enc_input, _enc_mask):
                outs = []
                out = start
                for i in range(max_length):
                    mask = to_tensor([True])
                    mask = tf.expand_dims(mask, axis=0)
                    mask = tf.tile(mask, [batch_size, 1])
                    out = stack((out, _enc_input), training=None, mask=(mask, _enc_mask), cache=cache,
                                decode_loop_step=i)
                    outs.append(out)
                outs = tf.concat(outs, axis=1)
                return outs

            ret = inc_encode(dec_input, encoded_out, enc_mask)
            check_input = tf.concat([dec_input, ret[:, :-1, :]], axis=1)
            # Call sanity check outside of cache_context to make sure we're getting the same-ish result
        sanity_check = stack((check_input, encoded_out), training=None, mask=(
        tf.cast(tf.ones((batch_size, max_length)), tf.bool), enc_mask))  # , cache=cache, decode_loop_step=0)

        print(tf.reduce_max(ret - sanity_check), tf.reduce_mean(ret - sanity_check))
        self.assertAllClose(ret, sanity_check, atol=3e-6, rtol=0.2)



import os

import tensorflow as tf
from tensorflow.python.framework import test_util

from dynastes.data_augmentation.audio_spectrum import spec_augment
from dynastes.ops import spectral_ops


class SpecAugmentTest(tf.test.TestCase):

    @test_util.use_deterministic_cudnn
    def test_simple(self):
        n_fft = 256
        hop_length = 64
        pad = 256
        dir_path = os.path.dirname(os.path.realpath(__file__))

        raw_audio = tf.io.read_file(dir_path + '/data/test_wav.wav')
        waves = tf.audio.decode_wav(raw_audio)[0][:20480] * 0.99
        waves.shape
        waves = tf.expand_dims(waves, 0)

        print(waves.shape)

        @tf.function
        def graph_test(waves):
            stfts = spectral_ops.waves_to_stfts(waves, n_fft=n_fft, hop_length=hop_length, pad_l=pad, pad_r=pad,
                                                discard_dc=True, hq=False)
            return stfts
        stfts = graph_test(waves)


        stfts_mags = tf.maximum(tf.minimum(tf.abs(stfts), 0.5), -0.5)
        stfts_angles = tf.math.angle(stfts)
        stfts = spectral_ops.polar2rect(stfts_mags, stfts_angles)

        l2mel = tf.signal.linear_to_mel_weight_matrix(num_spectrogram_bins=n_fft // 2,
                                                      sample_rate=44100,
                                                      lower_edge_hertz=50.,
                                                      upper_edge_hertz=6000.,
                                                      num_mel_bins=80)

        melmags = spectral_ops.stfts_to_melspecgrams(stfts, l2mel=l2mel, ifreq=False, return_phase=False)
        # melmags = tf.reduce_sum(melmags, axis=-1)
        print(melmags.shape)

        augs = []
        print(tf.reduce_max(melmags), tf.reduce_min(melmags))

        for i in range(50):
            augs.append(spec_augment(melmags))

        mean_augs = tf.reduce_mean(tf.stack(augs), axis=0)

        diff = tf.reduce_mean(melmags - mean_augs)
        print(diff)



import timeit

import numpy as np
import tensorflow as tf
from tensorflow.keras.utils import custom_object_scope
from tensorflow.python.framework import test_util

import dynastes as d
from dynastes.layers.t2t_attention_layers import Attention1D, Attention2D, PseudoBlockSparseAttention1D
from dynastes.probability.pseudoblocksparse_bijectors import BlockSparseStridedRoll1D


def _test_grads(testCase: tf.test.TestCase, func, input):
    _, grads = tf.test.compute_gradient(func, input)
    for grad in grads:
        testCase.assertNotAllClose(grad, np.zeros_like(grad))
        testCase.assertAllInRange(grad, -400., 400)


to_tensor = tf.convert_to_tensor
normal = np.random.normal


class T2TAttention1DTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        tf.config.optimizer.set_jit(True)
        t_steps = 12
        bs = 2
        dim = 16
        dim_mq = 32
        num_heads = 4
        s = 2

        layers = [

            (
                'PsuedoBlockSparse Masked',
                PseudoBlockSparseAttention1D(num_heads=num_heads, block_size=8,
                                             blocksparse_bijector=BlockSparseStridedRoll1D(block_size=8),
                                             mask_right=True),
                {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}),
            (
                'PsuedoBlockSparse Multiquery Masked',
                PseudoBlockSparseAttention1D(num_heads=num_heads, block_size=8,
                                             multiquery_attention=True,
                                             blocksparse_bijector=BlockSparseStridedRoll1D(block_size=8),
                                             mask_right=True),
                {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                 'dim_v': dim_mq // num_heads}),
            (
                'Local Masked',
                Attention1D(num_heads=num_heads, self_attention=True, local=True, masked=True, block_length=8,
                            filter_width=6),
                {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}),
            (
                'Local Multiquery Masked',
                Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=True, local=True,
                            masked=True, block_length=8,
                            filter_width=6),
                {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                 'dim_v': dim_mq // num_heads}),
            (
                'Sparse Unmasked',
                Attention1D(num_heads=num_heads, self_attention=True, sparse=True, lsh_bucket_length=3,
                            mask_right=True),
                {'self': True, 'steps_q': 32, 'steps_kv': 32, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}),
            (
                'Normal',
                Attention1D(num_heads=num_heads, self_attention=False),
                {'self': False, 'steps_q': 16, 'steps_kv': 64, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}
            ),
            (
                'Normal Multiquery',
                Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=False),
                {'self': False, 'steps_q': 16, 'steps_kv': 64, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                 'dim_v': dim_mq // num_heads}
            ),
            (
                'Relative',
                Attention1D(num_heads=num_heads, self_attention=True, relative=True, max_relative_position=8,
                            mask_right=True),
                {'self': True, 'steps_q': 16, 'steps_kv': 16, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}
            ),
            (
                'Relative Multiquery',
                Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=True, relative=True,
                            max_relative_position=8,
                            mask_right=True),
                {'self': True, 'steps_q': 32, 'steps_kv': 32, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                 'dim_v': dim_mq // num_heads}
            ),
            (
                'Relative Masked',
                Attention1D(num_heads=num_heads, self_attention=True, relative=True, max_relative_position=8,
                            masked=True,
                            mask_right=True),
                {'self': True, 'steps_q': 16, 'steps_kv': 16, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}
            ),
            (
                'Relative Multiquery Masked',
                Attention1D(num_heads=num_heads, multiquery_attention=True,
                            self_attention=True, relative=True, max_relative_position=8,
                            masked=True,
                            mask_right=True),
                {'self': True, 'steps_q': 16, 'steps_kv': 16, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                 'dim_v': dim_mq // num_heads}
            ),
            (
                'Local',
                Attention1D(num_heads=num_heads, self_attention=True, local=True, block_length=8, filter_width=6),
                {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}),
            (
                'Local Multiquery',
                Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=True, local=True,
                            block_length=8, filter_width=6),
                {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                 'dim_v': dim_mq // num_heads}),

        ]

        def test_layer(layer, params):
            q = to_tensor(normal(size=(bs, params['steps_q'], params['dim_q']))
                          .astype(np.float32))
            k = to_tensor(normal(size=(bs, params['steps_kv'], params['dim_k']))
                          .astype(np.float32))
            v = to_tensor(normal(size=(bs, params['steps_kv'], params['dim_v']))
                          .astype(np.float32))
            mask_len = params['steps_q'] // 6
            mask_q = to_tensor(([True] * (params['steps_q'] - mask_len)) + ([False] * (mask_len)))
            mask_q = tf.expand_dims(mask_q, axis=0)
            mask_q = tf.tile(mask_q, [bs, 1])

            mask_kv = to_tensor(([True] * (params['steps_kv'] - mask_len)) + ([False] * (mask_len)))
            mask_kv = tf.expand_dims(mask_kv, axis=0)
            mask_kv = tf.tile(mask_kv, [bs, 1])

            def get_test_fn(layer, mask):
                @tf.function
                def test_func(q, k, v):
                    r, _ = layer([q, k, v], mask=mask)
                    return r

                return test_func

            mask = [mask_q, mask_kv, mask_kv]
            r, _ = layer([q, k, v], mask=mask)

            test_fn = get_test_fn(layer, mask)
            # Warmup
            rfn = test_fn(q, k=k, v=v)
            self.assertAllClose(r, rfn)
            self.assertAllInRange(r, -200., 200)

            def fn():
                test_fn(q, k=k, v=v)

            time = timeit.timeit(fn, number=2) / (params['steps_q'] * params['steps_kv'] * params['dim_q'])
            time *= 8192

            _test_grads(self, test_fn, [q, k, v])
            return time

        for (type, layer, params) in layers:
            print(type, test_layer(layer, params))


class T2TAttention2DTest(tf.test.TestCase):
    @test_util.use_deterministic_cudnn
    def test_simple(self):
        tf.config.optimizer.set_jit(True)
        t_steps = 12
        bs = 2
        dim = 8
        dim_mq = 16
        num_heads = 4
        s = 2

        layers = [

            (
                'Relative',
                Attention2D(num_heads=num_heads, self_attention=True, relative=True, max_relative_position=8,
                            mask_right=True),
                {'self': True, 'steps_q': 8, 'steps_kv': 8, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}
            ),
            (
                'Relative Multiquery',
                Attention2D(num_heads=num_heads, multiquery_attention=True, self_attention=True, relative=True,
                            max_relative_position=8,
                            mask_right=True),
                {'self': True, 'steps_q': 8, 'steps_kv': 8, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
                 'dim_v': dim_mq // num_heads}
            )
        ]

        # TODO: Add these back but for 2D
        """
        (
            'Local',
            Attention1D(num_heads=num_heads, self_attention=True, local=True, block_length=8, filter_width=6),
            {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}),
        (
            'Local Multiquery',
            Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=True, local=True,
                        block_length=8, filter_width=6),
            {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
             'dim_v': dim_mq // num_heads}),
        (
            'Local Masked',
            Attention1D(num_heads=num_heads, self_attention=True, local=True, masked=True, block_length=8,
                        filter_width=6),
            {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim, 'dim_k': dim, 'dim_v': dim}),
        (
            'Local Multiquery Masked',
            Attention1D(num_heads=num_heads, multiquery_attention=True, self_attention=True, local=True,
                        masked=True, block_length=8,
                        filter_width=6),
            {'self': True, 'steps_q': 64, 'steps_kv': 64, 'dim_q': dim_mq, 'dim_k': dim_mq // num_heads,
             'dim_v': dim_mq // num_heads}),
         """

        def test_layer(layer, params):
            q = to_tensor(normal(size=(bs, params['steps_q'], params['steps_q'], params['dim_q']))
                          .astype(np.float32))
            k = to_tensor(normal(size=(bs, params['steps_kv'], params['steps_kv'], params['dim_k']))
                          .astype(np.float32))
            v = to_tensor(normal(size=(bs, params['steps_kv'], params['steps_kv'], params['dim_v']))
                          .astype(np.float32))

            # mask_q = to_tensor(([True] * (params['steps_q'] - 3)) + ([False] * (3)))
            # mask_q = tf.expand_dims(mask_q, axis=0)
            # mask_q = tf.tile(mask_q, [bs, 1])

            # mask_kv = to_tensor(([True] * (params['steps_kv'] - 3)) + ([False] * (3)))
            # mask_kv = tf.expand_dims(mask_kv, axis=0)
            # mask_kv = tf.tile(mask_kv, [bs, 1])

            def get_test_fn(layer, mask):
                @tf.function
                def test_func(q, k, v):
                    r, _ = layer([q, k, v], mask=mask)
                    return r

                return test_func

            # mask = [mask_q, mask_kv, mask_kv]
            mask = None
            r, _ = layer([q, k, v], mask=mask)
            test_fn = get_test_fn(layer, mask)
            # Warmup
            rfn = test_fn(q, k=k, v=v)
            self.assertAllClose(r, rfn)
            self.assertAllInRange(r, -200., 200)

            def fn():
                test_fn(q, k=k, v=v)

            time = timeit.timeit(fn, number=2) / (
                    params['steps_q'] * params['steps_q'] * params['steps_kv'] * params['steps_kv'] * params[
                'dim_q'])
            time *= 8192

            _test_grads(self, test_fn, [q, k, v])
            return time

        for (type, layer, params) in layers:
            print(type, test_layer(layer, params))


if __name__ == '__main__':
    tf.config.optimizer.set_jit(True)
    tf.test.main()



import abc
import copy

import tensorflow as tf
import tensorflow_probability as tfp

from dynastes.layers.base_layers import DynastesBaseLayer
from dynastes.ops.t2t_common import shape_list
from dynastes.probability import bijector_partials
from dynastes.probability.bijectors import EventShapeAwareChain

tfpb = tfp.bijectors


class PseudoBlockSparseBijector(DynastesBaseLayer, abc.ABC):

    @abc.abstractmethod
    def get_bijector(self, x) -> tfpb.Bijector:
        pass

    @abc.abstractmethod
    def get_causality_matrix(self, x) -> tf.Tensor:
        pass


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class PseudoBlockSparseBijector1D(PseudoBlockSparseBijector, abc.ABC):

    def get_causality_matrix(self, x):
        _, length, _ = shape_list(x)
        lrange = tf.range(0, length)
        return tf.reshape(lrange, [1, length, 1])


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class ChainedPseudoBlockSparseBijector1D(PseudoBlockSparseBijector1D):

    def __init__(self, partial_bijectors, **kwargs):
        super(ChainedPseudoBlockSparseBijector1D, self).__init__(**kwargs)
        self.partial_bijectors = partial_bijectors

    def get_bijector(self, x):
        event_shape_in = shape_list(x)[1:]
        chain = EventShapeAwareChain(event_shape_in, copy.copy(self.partial_bijectors))
        return chain

    def get_causality_matrix(self, x):
        lrange = super(ChainedPseudoBlockSparseBijector1D, self).get_causality_matrix(x)
        return self.get_bijector(lrange).forward(lrange)


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class BlockSparseStridedRoll1D(ChainedPseudoBlockSparseBijector1D):

    def __init__(self,
                 block_size,
                 stride=1,
                 partial_bijectors=None,
                 **kwargs):
        self.block_size = block_size
        self.stride = stride
        partial_bijectors = [
            bijector_partials.Reshape([-1, block_size, -1]),
            bijector_partials.RollIncremental(axis=1, steps=-(block_size + stride), constant=0),
            bijector_partials.Transpose(perm=[1, 0, 2]),
            bijector_partials.Reshape([-1, -1])
        ]
        super(BlockSparseStridedRoll1D, self).__init__(partial_bijectors=partial_bijectors, **kwargs)

    def get_config(self):
        config = {
            'block_size': self.block_size,
            'stride': self.stride,
        }
        base_config = super(BlockSparseStridedRoll1D, self).get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class BlockSparsePrimedStridedRoll1D(ChainedPseudoBlockSparseBijector1D):

    def __init__(self,
                 block_size,
                 n_primes=-1,
                 stride=1,
                 partial_bijectors=None,
                 **kwargs):

        self.block_size = block_size
        self.n_primes = n_primes
        self.stride = stride

        def prime_factors(n):
            i = 2
            factors = []
            while i * i <= n:
                if n % i:
                    i += 1
                else:
                    n //= i
                    factors.append(i)
            if n > 1:
                factors.append(n)
            return factors

        primes = prime_factors(block_size)
        if n_primes != -1:
            primes = primes[:n_primes - 1] + [sum(primes[n_primes - 1:])]
        else:
            n_primes = len(primes)

        partial_bijectors = [
            bijector_partials.Reshape([-1] + primes + [-1])
        ]
        for i in range(n_primes):
            partial_bijectors.append(
                bijector_partials.RollIncremental(axis=1 + i, steps=-(primes[i] + stride), constant=0)
            )
        partial_bijectors.append(
            bijector_partials.Transpose(perm=[0] + list(reversed(range(1, n_primes))) + [n_primes])
        )
        partial_bijectors.append(
            bijector_partials.Reshape([-1, -1])
        )
        super(BlockSparsePrimedStridedRoll1D, self).__init__(partial_bijectors=partial_bijectors, **kwargs)

    def get_config(self):
        config = {
            'block_size': self.block_size,
            'stride': self.stride,
            'n_primes': self.n_primes,
        }
        base_config = super(BlockSparsePrimedStridedRoll1D, self).get_config()
        return {**base_config, **config}



import math

import tensorflow as tf

from dynastes.util import t2t_expert_util
from dynastes.ops import t2t_common

def split_last_dimension(x, n):
    """Reshape x so that the last dimension becomes two dimensions.
    The first of these two dimensions is n.
    Args:
      x: a Tensor with shape [..., m]
      n: an integer.
    Returns:
      a Tensor with shape [..., n, m/n]
    """
    x_shape = t2t_common.shape_list(x)
    m = x_shape[-1]
    if isinstance(m, int) and isinstance(n, int):
        assert m % n == 0
    return tf.reshape(x, x_shape[:-1] + [n, m // n])


def combine_last_two_dimensions(x):
    """Reshape x so that the last two dimension become one.
    Args:
      x: a Tensor with shape [..., a, b]
    Returns:
      a Tensor with shape [..., ab]
    """
    x_shape = t2t_common.shape_list(x)
    a, b = x_shape[-2:]
    return tf.reshape(x, x_shape[:-2] + [a * b])


def combine_first_two_dimensions(x):
    """Reshape x so that the first two dimension become one.
    Args:
      x: a Tensor with shape [a, b, ...]
    Returns:
      a Tensor with shape [ab, ...]
    """
    ret = tf.reshape(x, tf.concat([[-1], t2t_common.shape_list(x)[2:]], 0))
    old_shape = x.get_shape().dims
    a, b = old_shape[:2]
    new_shape = [a * b if a and b else None] + old_shape[2:]
    ret.set_shape(new_shape)
    return ret


def split_heads(x, num_heads):
    """Split channels (dimension 2) into multiple heads (becomes dimension 1).
    Args:
      x: a Tensor with shape [batch, length, channels]
      num_heads: an integer
    Returns:
      a Tensor with shape [batch, num_heads, length, channels / num_heads]
    """
    return tf.transpose(split_last_dimension(x, num_heads), [0, 2, 1, 3])


def split_heads_2d(x, num_heads):
    """Split channels (dimension 3) into multiple heads (becomes dimension 1).
    Args:
      x: a Tensor with shape [batch, height, width, channels]
      num_heads: an integer
    Returns:
      a Tensor with shape [batch, num_heads, height, width, channels / num_heads]
    """
    return tf.transpose(split_last_dimension(x, num_heads), [0, 3, 1, 2, 4])


def split_heads_nd(x, num_heads):
    """Split the depth dimension (last dimension) into multiple heads.
    Args:
      x: a [batch, d1, ..., dn, depth] tensor
      num_heads: an integer
    Returns:
      a [batch, num_heads, d1, ..., dn, depth // num_heads]
    """
    num_dimensions = len(t2t_common.shape_list(x)) - 2
    return tf.transpose(
        split_last_dimension(x, num_heads), [0, num_dimensions + 1] +
                                            list(range(1, num_dimensions + 1)) + [num_dimensions + 2])


def combine_heads(x):
    """Inverse of split_heads.
    Args:
      x: a Tensor with shape [batch, num_heads, length, channels / num_heads]
    Returns:
      a Tensor with shape [batch, length, channels]
    """
    return combine_last_two_dimensions(tf.transpose(x, [0, 2, 1, 3]))


def combine_heads_2d(x):
    """Inverse of split_heads_2d.
    Args:
      x: a Tensor with shape
        [batch, num_heads, height, width, channels / num_heads]
    Returns:
      a Tensor with shape [batch, height, width, channels]
    """
    return combine_last_two_dimensions(tf.transpose(x, [0, 2, 3, 1, 4]))


def combine_heads_nd(x):
    """Inverse of split_heads_nd.
    Args:
      x: a [batch, num_heads, d1, ..., dn, depth // num_heads] tensor
    Returns:
      a [batch, d1, ...., dn, depth] tensor
    """
    num_dimensions = len(t2t_common.shape_list(x)) - 3
    return combine_last_two_dimensions(
        tf.transpose(x, [0] + list(range(2, num_dimensions + 2)) +
                     [1, num_dimensions + 2]))


def to_int32(x):
    """Cast x to float; created because to_float is deprecated."""
    return tf.cast(x, tf.int32)


def mixed_precision_is_enabled(
        activation_dtype=None, weight_dtype=None, hparams=None):
    assert not (hparams and (activation_dtype or weight_dtype)), (
        "Provide only hparams or activation_dtype and weight_dtype")
    if (hparams and hasattr(hparams, "activation_dtype") and
            hasattr(hparams, "weight_dtype")):
        activation_dtype = hparams.activation_dtype
        weight_dtype = hparams.weight_dtype
    return activation_dtype == tf.float16 and weight_dtype == tf.float32


def maybe_upcast(logits,
                 activation_dtype=None, weight_dtype=None, hparams=None):
    if mixed_precision_is_enabled(activation_dtype, weight_dtype, hparams):
        return tf.cast(logits, tf.float32)
    return logits


def harden_attention_weights(weights, k, gumbel_noise_weight):
    """Make attention weights non-0 only on the top k ones."""
    if gumbel_noise_weight > 0.:
        gumbel_noise = -tf.log(-tf.log(tf.random_uniform(tf.shape(weights),
                                                         minval=1e-5,
                                                         maxval=1 - 1e-5)))
        weights += gumbel_noise * gumbel_noise_weight

    # Subtract the top-kth weight and zero-out all lower ones.
    # Note that currently in case of numerical ties it will retain more
    # than k elements. In the future, we may want to avoid this.
    weights -= t2t_common.top_kth_iterative(weights, k)
    weights = tf.nn.relu(weights)
    # Re-normalize the weights.
    weights_sum = tf.reduce_sum(weights, axis=-1, keep_dims=True)
    weights_sum = tf.maximum(weights_sum, 1e-6)  # Avoid division by 0.
    weights /= weights_sum
    return weights


def dot_product_attention(q,
                          k,
                          v,
                          bias,
                          dropout_rate=0.0,
                          name='dot_product_attention',
                          dropout_broadcast_dims=[0,1],
                          activation_dtype=None,
                          weight_dtype=None,
                          hard_attention_k=0,
                          gumbel_noise_weight=0.0):
    """Dot-product attention.
    Args:
      q: Tensor with shape [..., length_q, depth_k].
      k: Tensor with shape [..., length_kv, depth_k]. Leading dimensions must
        match with q.
      v: Tensor with shape [..., length_kv, depth_v] Leading dimensions must
        match with q.
      bias: bias Tensor (see attention_bias())
      dropout_rate: a float.
      save_weights_to: an optional dictionary to capture attention weights
        for visualization; the weights tensor will be appended there under
        a string key created from the variable scope (including name).
      dropout_broadcast_dims: an optional list of integers less than rank of q.
        Specifies in which dimensions to broadcast the dropout decisions.
      activation_dtype: Used to define function activation dtype when using
        mixed precision.
      weight_dtype: The dtype weights are stored in when using mixed precision
      hard_attention_k: integer, if > 0 triggers hard attention (picking top-k)
      gumbel_noise_weight: if > 0, apply Gumbel noise with weight
        `gumbel_noise_weight` before picking top-k. This is a no op if
        hard_attention_k <= 0.
    Returns:
      Tensor with shape [..., length_q, depth_v].
    """

    logits = tf.matmul(q, k, transpose_b=True)  # [..., length_q, length_kv]
    if bias is not None:
        bias = t2t_common.cast_like(bias, logits)
        logits += bias
    # If logits are fp16, upcast before softmax
    logits = maybe_upcast(logits, activation_dtype, weight_dtype)
    weights = tf.nn.softmax(logits, name="attention_weights")
    if hard_attention_k > 0:
        weights = harden_attention_weights(weights, hard_attention_k,
                                           gumbel_noise_weight)
    weights = t2t_common.cast_like(weights, q)
    # Drop out attention links for each head.
    weights = t2t_common.dropout_with_broadcast_dims(
        weights, rate=dropout_rate, broadcast_dims=dropout_broadcast_dims)
    return tf.matmul(weights, v), {name + '/attention_weights': weights}


def get_embedding_initializer_stddev(depth):
    return depth ** -0.5


def get_relative_embeddings_left_right_shape(max_relative_position, depth,
                                             num_heads,
                                             heads_share_relative_embedding):
    max_relative_position_unmasked = 2 * max_relative_position - 1
    if heads_share_relative_embedding:
        embedding_shape = (max_relative_position_unmasked, depth)
    else:
        embedding_shape = (num_heads, max_relative_position_unmasked, depth)
    return embedding_shape


def get_relative_embeddings_left_shape(max_relative_position, depth, num_heads, heads_share_relative_embedding):
    if heads_share_relative_embedding:
        embedding_shape = (max_relative_position, depth)
    else:
        embedding_shape = (num_heads, max_relative_position, depth)
    return embedding_shape


def get_relative_embeddings_left_right(relative_embeddings,
                                       max_relative_position, length, depth,
                                       heads_share_relative_embedding):
    """Instantiate or retrieve relative embeddings, sliced according to length.
    Use for unmasked case where the relative attention looks both left and right.
    Args:
      max_relative_position: an Integer for the number of entries in the relative
        embedding, which corresponds to the max relative distance that is
        considered.
      length: an Integer, specifies the length of the input sequence for which
        this relative embedding is retrieved for.
      depth: an Integer, specifies the depth for relative embeddings.
      num_heads: an Integer, specifies the number of heads.
      heads_share_relative_embedding: a Boolean specifying if the relative
        embedding is shared across heads.
      name: a string giving the name of the embedding variables.
    Returns:
      a Tensor with shape [length, depth]
    """
    # Pad first before slice to avoid using tf.cond.
    pad_length = tf.maximum(length - max_relative_position, 0)
    slice_start_position = tf.maximum(max_relative_position - length, 0)
    if heads_share_relative_embedding:
        padded_relative_embeddings = tf.pad(
            relative_embeddings,
            [[pad_length, pad_length], [0, 0]])
        used_relative_embeddings = tf.slice(
            padded_relative_embeddings,
            [slice_start_position, 0], [2 * length - 1, -1])
    else:
        padded_relative_embeddings = tf.pad(
            relative_embeddings,
            [[0, 0], [pad_length, pad_length], [0, 0]])
        used_relative_embeddings = tf.slice(
            padded_relative_embeddings,
            [0, slice_start_position, 0], [-1, 2 * length - 1, -1])
    return used_relative_embeddings


def get_relative_embeddings_left(relative_embeddings, max_relative_position, length, heads_share_relative_embedding):
    """Instantiate or retrieve relative embeddings, sliced according to length.
    Use for masked case where the relative attention is only looking left.
    Args:
      max_relative_position: an Integer for the number of entries in the relative
        embedding, which corresponds to the max relative distance that is
        considered.
      length: an Integer, specifies the length of the input sequence for which
        this relative embedding is retrieved for.
      depth: an Integer, specifies the depth for relative embeddings.
      num_heads: an Integer, specifies the number of heads.
      heads_share_relative_embedding: a Boolean specifying if the relative
        embedding is shared across heads.
      name: a string giving the name of the embedding variables.
    Returns:
      a Tensor with shape [length, depth]
    """
    # Pad first before slice to avoid using tf.cond.
    pad_length = tf.maximum(length - max_relative_position, 0)
    start_slice_position = tf.maximum(max_relative_position - length, 0)
    if heads_share_relative_embedding:
        padded_relative_embeddings = tf.pad(
            relative_embeddings,
            [[pad_length, 0], [0, 0]])
        used_relative_embeddings = tf.slice(
            padded_relative_embeddings,
            [start_slice_position, 0], [length, -1])
    else:
        padded_relative_embeddings = tf.pad(
            relative_embeddings,
            [[0, 0], [pad_length, 0], [0, 0]])
        used_relative_embeddings = tf.slice(
            padded_relative_embeddings,
            [0, start_slice_position, 0], [-1, length, -1])
    return used_relative_embeddings


def _absolute_position_to_relative_position_unmasked(x):
    """Helper function for dot_product_unmasked_self_attention_relative_v2.
    Rearrange an attention logits or weights Tensor.
    The dimensions of the input represent:
    [batch, heads, query_position, memory_position]
    The dimensions of the output represent:
    [batch, heads, query_position, memory_position - query_position + length - 1]
    Only works with unmasked_attention.
    Args:
      x: a Tensor with shape [batch, heads, length, length]
    Returns:
      a Tensor with shape [batch, heads, length, 2*length-1]
    """
    batch, heads, length, _ = t2t_common.shape_list(x)
    # padd along column
    x = tf.pad(x, [[0, 0], [0, 0], [0, 0], [0, length - 1]])
    x_flat = tf.reshape(x, [batch, heads, length ** 2 + length * (length - 1)])
    # add 0's in the beginning that will skew the elements after reshape
    x_flat = tf.pad(x_flat, [[0, 0], [0, 0], [length, 0]])
    x = tf.reshape(x_flat, [batch, heads, length, 2 * length])
    x = tf.slice(x, [0, 0, 0, 1], [batch, heads, length,
                                   2 * length - 1])
    return x


def _relative_position_to_absolute_position_unmasked(x):
    """Converts tensor from relative to aboslute indexing for local attention.
    Args:
      x: a Tensor of shape [batch (or batch*num_blocks), heads,
                            length, 2 * length - 1]
    Returns:
      A Tensor of shape [batch (or batch*num_blocks), heads, length, length]
    """
    x_shape = t2t_common.shape_list(x)
    batch = x_shape[0]
    heads = x_shape[1]
    length = x_shape[2]
    # Concat columns of pad to shift from relative to absolute indexing.
    col_pad = tf.zeros((batch, heads, length, 1))
    x = tf.concat([x, col_pad], axis=3)

    # Concat extra elements so to add up to shape (len+1, 2*len-1).
    flat_x = tf.reshape(x, [batch, heads, length * 2 * length])
    flat_pad = tf.zeros((batch, heads, length - 1))
    flat_x_padded = tf.concat([flat_x, flat_pad], axis=2)

    # Reshape and slice out the padded elements.
    final_x = tf.reshape(flat_x_padded, [batch, heads, length + 1, 2 * length - 1])
    final_x = final_x[:, :, :, length - 1:]
    final_x = final_x[:, :, :length, :]
    return final_x


def _absolute_position_to_relative_position_masked(x):
    """Helper to dot_product_self_attention_relative_v2.
    Rearrange an attention logits or weights Tensor.
    The dimensions of the input represent:
    [batch, heads, query_position, memory_position]
    The dimensions of the output represent:
    [batch, heads, query_position, memory_position - query_position + length - 1]
    Only works with masked_attention.  Undefined behavior for regions of the
    input where memory_position > query_position.
    Args:
      x: a Tensor with shape [batch, heads, length, length]
    Returns:
      a Tensor with shape [batch, heads, length, length]
    """
    batch, heads, length, _ = t2t_common.shape_list(x)
    x = tf.pad(x, [[0, 0], [0, 0], [1, 0], [0, 0]])
    x = tf.reshape(x, [batch, heads, length, length + 1])
    x = tf.slice(x, [0, 0, 0, 1], [batch, heads, length, length])
    return x


def _relative_position_to_absolute_position_masked(x):
    """Helper to dot_product_self_attention_relative_v2.
    Rearrange an attention logits or weights Tensor.
    The dimensions of the input represent:
    [batch, heads, query_position, memory_position - query_position + length - 1]
    The dimensions of the output represent:
    [batch, heads, query_position, memory_position]
    Only works with masked_attention.  Undefined behavior for regions of the
    input where memory_position > query_position.
    Args:
      x: a Tensor with shape [batch, heads, length, length]
    Returns:
      a Tensor with shape [batch, heads, length, length]
    """
    batch, heads, length, _ = t2t_common.shape_list(x)
    x = tf.pad(x, [[0, 0], [0, 0], [0, 0], [1, 0]])
    x = tf.reshape(x, [batch, heads, 1 + length, length])
    x = tf.slice(x, [0, 0, 1, 0], [-1, -1, -1, -1])
    return x


def matmul_with_relative_values(x, y, heads_share_relative_embedding):
    if heads_share_relative_embedding:
        ret = tf.einsum("bhlm,md->bhld", x, y)
    else:
        ret = tf.einsum("bhlm,hmd->bhld", x, y)
    return ret


def matmul_with_relative_keys(x, y, heads_share_relative_embedding):
    if heads_share_relative_embedding:
        ret = tf.einsum("bhld,md->bhlm", x, y)
    else:
        ret = tf.einsum("bhld,hmd->bhlm", x, y)
    return ret


def dot_product_unmasked_self_attention_relative_v2(
        q, k, v, bias, key_leftright_embeddings, value_leftright_embeddings=None,
        max_relative_position=None, dropout_rate=0.0, save_weights_to=None,
        name='dot_product_unmasked_self_attention_relative_v2',
        dropout_broadcast_dims=[0,1], heads_share_relative_embedding=False,
        add_relative_to_values=False):
    """Calculate relative position-aware dot-product self-attention.
    The attention calculation is augmented with learned representations for the
    relative position between each element in q and each element in k and v.
    Args:
      q: a Tensor with shape [batch, heads, length, depth].
      k: a Tensor with shape [batch, heads, length, depth].
      v: a Tensor with shape [batch, heads, length, depth].
      bias: bias Tensor.
      max_relative_position: an integer the max relative embedding considered.
        Changing this invalidates checkpoints.
      dropout_rate: a floating point number.
      save_weights_to: an optional dictionary to capture attention weights
        for visualization; the weights tensor will be appended there under
        a string key created from the variable scope (including name).
      name: an optional string.
      dropout_broadcast_dims:  an optional list of integers less than 4
        specifying in which dimensions to broadcast the dropout decisions.
        saves memory.
      heads_share_relative_embedding: a boolean indicating wheather to share
        relative embeddings between attention heads.
      add_relative_to_values: a boolean for whether to add relative component to
        values.
    Returns:
      A Tensor.
    Raises:
      ValueError: if max_relative_position is not > 0.
    """
    if not max_relative_position:
        raise ValueError("Max relative position (%s) should be > 0 when using "
                         "relative self attention." % (max_relative_position))

    # This calculation only works for self attention.
    # q, k and v must therefore have the same shape.
    q.get_shape()[2:].assert_is_compatible_with(k.get_shape()[2:])
    q.get_shape()[2:-1].assert_is_compatible_with(v.get_shape()[2:-1])

    # [batch, num_heads, query_length, memory_length]
    logits = tf.matmul(q, k, transpose_b=True)

    # Use separate embeddings suitable for keys and values.
    _, q_heads, length, _ = t2t_common.shape_list(q)
    _, k_heads, _, depth_k = t2t_common.shape_list(k)
    mqa = k_heads == 1 and q_heads > k_heads

    key_relative_embeddings = get_relative_embeddings_left_right(
        key_leftright_embeddings,
        max_relative_position, length, depth_k,
        heads_share_relative_embedding)
    unmasked_rel_logits = matmul_with_relative_keys(
        q, key_relative_embeddings, heads_share_relative_embedding)
    unmasked_rel_logits = _relative_position_to_absolute_position_unmasked(
        unmasked_rel_logits)
    logits += unmasked_rel_logits

    if bias is not None:
        logits += bias

    weights = tf.nn.softmax(logits, name="attention_weights")
    # dropping out the attention links for each of the heads
    weights = t2t_common.dropout_with_broadcast_dims(
        weights, rate=dropout_rate, broadcast_dims=dropout_broadcast_dims)
    # relative_weights.set_shape([None, None, None, max_length])
    ret = tf.matmul(weights, v)
    if add_relative_to_values:
        # Adds the contribution of the weighted relative embeddings to the values.
        # [batch, num_heads, query_length, 2*memory_length-1]
        relative_weights = _absolute_position_to_relative_position_unmasked(
            weights)
        depth_v = t2t_common.shape_list(v)[3]
        value_relative_embeddings = get_relative_embeddings_left_right(
            value_leftright_embeddings,
            max_relative_position, length, depth_v, heads_share_relative_embedding)
        ret += matmul_with_relative_values(
            relative_weights, value_relative_embeddings,
            heads_share_relative_embedding)
    return ret, {name + '_attention_weights': weights}


def dot_product_self_attention_relative_v2(q,
                                           k,
                                           v,
                                           bias,
                                           key_left_embedding,
                                           value_left_embedding=None,
                                           max_relative_position=None,
                                           dropout_rate=0.0,
                                           save_weights_to=None,
                                           name='dot_product_self_attention_relative_v2',
                                           dropout_broadcast_dims=[0,1],
                                           heads_share_relative_embedding=False,
                                           add_relative_to_values=False):
    """Calculate relative position-aware dot-product self-attention.
    Only works for masked self-attention (no looking forward).
    The attention calculation is augmented with learned representations for the
    relative position between each element in q and each element in k and v.
    Args:
      q: a Tensor with shape [batch, heads, length, depth].
      k: a Tensor with shape [batch, heads, length, depth].
      v: a Tensor with shape [batch, heads, length, depth].
      bias: bias Tensor.
      max_relative_position: an integer indicating the maximum relative distance
        to look back - changing this invalidates checkpoints
      dropout_rate: a floating point number.
      save_weights_to: an optional dictionary to capture attention weights
        for visualization; the weights tensor will be appended there under
        a string key created from the variable scope (including name).
      name: an optional string.
      dropout_broadcast_dims:  an optional list of integers less than 4
        specifying in which dimensions to broadcast the dropout decisions.
        saves memory.
      heads_share_relative_embedding: a boolean indicating wheather to share
        relative embeddings between attention heads.
      add_relative_to_values: a boolean for whether to add relative component to
        values.
    Returns:
      A Tensor.
    Raises:
      ValueError: if max_relative_position is not > 0.
    """
    if not max_relative_position:
        raise ValueError("Max relative position (%s) should be > 0 when using "
                         "relative self attention." % (max_relative_position))

    # This calculation only works for self attention.
    # q, k and v must therefore have the same shape.
    # (Except v can have different depth.)
    #q.get_shape()[2:].assert_is_compatible_with(k.get_shape()[2:])
    #q.get_shape()[2:-1].assert_is_compatible_with(v.get_shape()[2:-1])

    # Use separate embeddings suitable for keys and values.
    _, q_heads, length, _ = t2t_common.shape_list(q)
    _, k_heads, _, _ = t2t_common.shape_list(k)
    mqa = k_heads == 1 and q_heads > k_heads

    # [batch, num_heads, query_length, memory_length]
    logits = tf.matmul(q, k, transpose_b=True)
    key_relative_embeddings = get_relative_embeddings_left(key_left_embedding, max_relative_position, length,
                                                           heads_share_relative_embedding)

    rel_logits = matmul_with_relative_keys(q, key_relative_embeddings,
                                           heads_share_relative_embedding)
    rel_logits = _relative_position_to_absolute_position_masked(rel_logits)
    logits += rel_logits
    if bias is not None:
        logits += bias

    weights = tf.nn.softmax(logits, name="attention_weights")
    # Dropping out the attention links for each of the heads.
    weights = t2t_common.dropout_with_broadcast_dims(
        weights, dropout_rate, broadcast_dims=dropout_broadcast_dims)
    output = tf.matmul(weights, v)
    if add_relative_to_values:
        # [batch, num_heads, query_length, memory_length]
        relative_weights = _absolute_position_to_relative_position_masked(weights)
        depth_v = t2t_common.shape_list(v)[3]
        value_relative_embeddings = get_relative_embeddings_left(value_left_embedding, max_relative_position,
                                                                 length,
                                                                 heads_share_relative_embedding)
        output += matmul_with_relative_values(
            relative_weights, value_relative_embeddings,
            heads_share_relative_embedding)
    return output, {name + '_attention_weights': weights}


def embedding_to_padding(emb):
    """Calculates the padding bias based on which embeddings are all zero.
    We have hacked symbol_modality to return all-zero embeddings for padding.
    Args:
      emb: a Tensor with shape [..., depth].
    Returns:
      a float Tensor with shape [...]. Each element is 1 if its corresponding
      embedding vector is all zero, and is 0 otherwise.
    """
    emb_sum = tf.reduce_sum(tf.abs(emb), axis=-1)
    return t2t_common.to_float(tf.equal(emb_sum, 0.0))


def reshape_by_blocks(x, x_shape, memory_block_size):
    """Reshapes input by splitting its length over blocks of memory_block_size.
    Args:
      x: a Tensor with shape [batch, heads, length, depth]
      x_shape: tf.TensorShape of x.
      memory_block_size: Integer which divides length.
    Returns:
      Tensor with shape
      [batch, heads, length // memory_block_size, memory_block_size, depth].
    """
    x = tf.reshape(x, [
        x_shape[0], x_shape[1], x_shape[2] // memory_block_size,
        memory_block_size, x_shape[3]
    ])
    return x


def attention_bias_local(length, max_backward, max_forward):
    """Create an bias tensor to be added to attention logits.
    A position may attend to positions at most max_distance from it,
    forward and backwards.
    This does not actually save any computation.
    Args:
      length: int
      max_backward: int, maximum distance backward to attend. Negative values
        indicate unlimited.
      max_forward: int, maximum distance forward to attend. Negative values
        indicate unlimited.
    Returns:
      a `Tensor` with shape [1, 1, length, length].
    """
    band = t2t_common.ones_matrix_band_part(
        length,
        length,
        max_backward,
        max_forward,
        out_shape=[1, 1, length, length])
    return -1e9 * (1.0 - band)


def attention_bias_lower_triangle(length):
    """Create an bias tensor to be added to attention logits.
    Allows a query to attend to all positions up to and including its own.
    Args:
     length: a Scalar.
    Returns:
      a `Tensor` with shape [1, 1, length, length].
    """
    return attention_bias_local(length, -1, 0)


def local_attention_1d(q, k, v, block_length=128, filter_width=100, name='local_attention_1d'):
    """Strided block local self-attention.
    The sequence is divided into blocks of length block_length. Attention for a
    given query position can see all memory positions in the corresponding block
    and filter_width many positions to the left and right of the block.
    Args:
      q: a Tensor with shape [batch, heads, length, depth_k]
      k: a Tensor with shape [batch, heads, length, depth_k]
      v: a Tensor with shape [batch, heads, length, depth_v]
      block_length: an integer
      filter_width: an integer indicating how much to look left and right of the
        block.
      name: an optional string
    Returns:
      a Tensor of shape [batch, heads, length, depth_v]
    """
    # Check that q, k, v have the same shape except in their depth dimension.
    q.get_shape()[2:-1].assert_is_compatible_with(k.get_shape()[2:-1])
    q.get_shape()[2:-1].assert_is_compatible_with(v.get_shape()[2:-1])

    batch_size, num_heads, original_length, _ = t2t_common.shape_list(q)

    # Pad query, key, value to ensure multiple of corresponding lengths.
    def pad_to_multiple(x, pad_length):
        x_length = t2t_common.shape_list(x)[2]
        return tf.pad(x, [[0, 0], [0, 0], [0, -x_length % pad_length], [0, 0]])

    def pad_l_and_r(x, pad_length):
        return tf.pad(x, [[0, 0], [0, 0], [pad_length, pad_length], [0, 0]])

    # Set up query blocks.
    # [batch, heads, blocks_q, block_length, depth_k]
    q = pad_to_multiple(q, block_length)
    q = reshape_by_blocks(q, t2t_common.shape_list(q), block_length)
    total_query_blocks = t2t_common.shape_list(q)[2]

    # Set up key and value blocks.
    # [batch, heads, blocks_k, block_length, depth_k]
    blocks_per_filter_width = filter_width // block_length
    remaining_items = filter_width % block_length
    k = pad_to_multiple(k, block_length)
    v = pad_to_multiple(v, block_length)
    k = pad_l_and_r(k, filter_width + block_length - remaining_items)
    v = pad_l_and_r(v, filter_width + block_length - remaining_items)
    k = reshape_by_blocks(k, t2t_common.shape_list(k), block_length)
    v = reshape_by_blocks(v, t2t_common.shape_list(v), block_length)

    total_kv_blocks = t2t_common.shape_list(k)[2]

    slices = []
    # prepare the left-most and right-most partial blocks if needed
    if remaining_items:
        first_partial_block_k = tf.slice(
            k, [0, 0, 0, block_length - remaining_items, 0],
            [-1, -1, total_query_blocks, -1, -1])
        first_partial_block_v = tf.slice(
            v, [0, 0, 0, block_length - remaining_items, 0],
            [-1, -1, total_query_blocks, -1, -1])
        last_partial_block_k = tf.slice(
            k, [0, 0, total_kv_blocks - total_query_blocks, 0, 0],
            [-1, -1, -1, remaining_items, -1])
        last_partial_block_v = tf.slice(
            v, [0, 0, total_kv_blocks - total_query_blocks, 0, 0],
            [-1, -1, -1, remaining_items, -1])
        slices.append((first_partial_block_k, first_partial_block_v))
        slices.append((last_partial_block_k, last_partial_block_v))

    # Prepare the rest of the blocks
    first_block_index = 1 if remaining_items else 0
    attention_blocks = 2 * blocks_per_filter_width + 1
    for i in range(first_block_index, attention_blocks + first_block_index):
        block_k = tf.slice(k, [0, 0, i, 0, 0],
                           [-1, -1, total_query_blocks, -1, -1])
        block_v = tf.slice(v, [0, 0, i, 0, 0],
                           [-1, -1, total_query_blocks, -1, -1])
        slices.append((block_k, block_v))
    # [batch, heads, blocks_q, block_length + 2 * filter_width, depth_k]
    k = tf.concat([s[0] for s in slices], axis=3)
    v = tf.concat([s[1] for s in slices], axis=3)

    attention_bias = tf.expand_dims(embedding_to_padding(k) * -1e9, axis=-2)
    depth_v = t2t_common.shape_list(v)[-1]

    output, _w_dict = dot_product_attention(
        q,
        k,
        v,
        bias=attention_bias,
        dropout_rate=0.,
        name="local_1d", )
    output = tf.reshape(output, [batch_size, num_heads, -1, depth_v])

    # Remove the padding if introduced.
    output = tf.slice(output, [0, 0, 0, 0], [-1, -1, original_length, -1])
    output.set_shape([None if isinstance(dim, tf.Tensor) else dim for dim in
                      (batch_size, num_heads, original_length, depth_v)])
    return output, _w_dict


def _make_local_block(x, depth, batch, heads, num_blocks, block_length):
    """Helper function to create a local version of the keys or values for 1d."""
    prev_block = tf.slice(x, [0, 0, 0, 0, 0],
                          [-1, -1, num_blocks - 1, -1, -1])
    cur_block = tf.slice(x, [0, 0, 1, 0, 0], [-1, -1, -1, -1, -1])
    local_block = tf.concat([prev_block, cur_block], 3)
    return tf.reshape(local_block,
                      [batch, heads, num_blocks - 1, block_length * 2, depth])


def masked_local_attention_1d(q,
                              k,
                              v,
                              mask=None,
                              causality_tensor=None,
                              mask_right=True,
                              block_length=128,
                              dropout_rate=0.,
                              name='masked_local_attention_1d'):
    """Attention to the source position and a neighborhood to the left of it.
    The sequence is divided into blocks of length block_length. Attention for a
    given query position can only see memory positions less than or equal to the
    query position, in the corresponding block and the previous block.
    Args:
      q: a Tensor with shape [batch, heads, length, depth_k]
      k: a Tensor with shape [batch, heads, length, depth_k]
      v: a Tensor with shape [batch, heads, length, depth_v]
      mask: optional mask of shape [batch, length] where 1 = attendable
      causality_tensor: optional tensor of shape [1/batch, length] enumerating position
      mask_right: creates right-masking if mask is not None if set to True (default: True)
      block_length: an integer
      dropout_rate: Dropout rate for attention dropout
      name: an optional string
    Returns:
      a Tensor of shape [batch, heads, length, depth_v]

    NOTE:
        USE OF MASK and CAUSALITY TENSOR IS EXPERIMENTAL!
    """

    batch, heads, length, depth_k = t2t_common.shape_list(q)
    _, heads_kv, _, _ = t2t_common.shape_list(k)
    depth_v = t2t_common.shape_list(v)[-1]
    if isinstance(block_length, tf.Tensor):
        const = tf.get_static_value(block_length)
        if const is not None:
            block_length = int(const)
    # If (length < 2 * block_length), then we use only one block.
    if isinstance(length, int) and isinstance(block_length, int):
        block_length = length if length < block_length * 2 else block_length
    else:
        block_length = tf.where(
            tf.less(length, block_length * 2), length, block_length)
    ct_batch = -1
    if causality_tensor is not None:
        ct_batch, _ = t2t_common.shape_list(causality_tensor)
    if causality_tensor is None and mask_right and mask is not None:
        causality_tensor = tf.range(0, length)
        causality_tensor = tf.reshape(causality_tensor, [1, length])
        ct_batch = 1
    # Pad query, key, value to ensure multiple of block length.
    original_length = length
    padding_size = tf.math.mod(-length, block_length)
    length += padding_size
    padding = [[0, 0], [0, 0], [0, padding_size], [0, 0]]
    q = tf.pad(q, padding)
    k = tf.pad(k, padding)
    v = tf.pad(v, padding)
    if mask is not None:
        # tf.get_logger().warning('Use of mask and/or causality tensor in masked_local_attention_1d is experimental')
        mask = tf.reshape(mask, [batch, 1, length, 1])
        mask = tf.pad(mask, padding)
        if mask_right or causality_tensor is not None:
            causality_tensor = tf.reshape(causality_tensor, [ct_batch, 1, length, 1])
            causality_tensor = tf.pad(causality_tensor, padding, constant_values=length + 1)
    if isinstance(length, int) and isinstance(block_length, int):
        num_blocks = length // block_length
    else:
        num_blocks = tf.cast(tf.math.divide(length, block_length), tf.int32)

    # Compute attention for the first query block.
    first_q = tf.slice(q, [0, 0, 0, 0], [-1, -1, block_length, -1])
    first_k = tf.slice(k, [0, 0, 0, 0], [-1, -1, block_length, -1])
    first_v = tf.slice(v, [0, 0, 0, 0], [-1, -1, block_length, -1])
    if mask is not None:
        first_mask = tf.slice(mask, [0, 0, 0, 0], [-1, -1, block_length, -1])
        if mask_right or causality_tensor is not None:
            first_ct = tf.slice(causality_tensor, [0, 0, 0, 0], [-1, -1, block_length, -1])
            first_ct = tf.cast(tf.linalg.matrix_transpose(first_ct) <= first_ct, mask.dtype)
            first_mask = tf.minimum(first_mask, first_ct)
        first_bias = -1e9 * (1.0 - first_mask)
    else:
        first_bias = attention_bias_lower_triangle(block_length)

    first_output, first_dict = dot_product_attention(
        first_q,
        first_k,
        first_v,
        bias=first_bias,
        dropout_rate=dropout_rate,
        name="first_block")

    # Compute attention for all subsequent query blocks.
    q = tf.reshape(q, [batch, heads, num_blocks, block_length, depth_k])
    k = tf.reshape(k, [batch, heads_kv, num_blocks, block_length, depth_k])
    v = tf.reshape(v, [batch, heads_kv, num_blocks, block_length, depth_v])
    if mask is not None:
        mask = tf.reshape(mask, [batch, 1, num_blocks, block_length, 1])
        local_mask = _make_local_block(mask, 1, batch, 1, num_blocks,
                                       block_length)
        tail_mask = tf.slice(mask, [0, 0, 1, 0, 0], [-1, -1, -1, -1, -1])
        tail_mask = tf.reshape(tail_mask,
                               [batch, 1, num_blocks - 1, block_length, 1])
        local_mask = tf.minimum(tf.linalg.matrix_transpose(local_mask), tail_mask)
        if causality_tensor is not None:
            ct = tf.reshape(causality_tensor, [ct_batch, 1, num_blocks, block_length, 1])
            local_ct = _make_local_block(ct, 1, ct_batch, 1, num_blocks,
                                         block_length)
            tail_ct = tf.slice(ct, [0, 0, 1, 0, 0], [-1, -1, -1, -1, -1])
            tail_ct = tf.reshape(tail_ct, [ct_batch, 1, num_blocks - 1, block_length, 1])
            ct_mask = tf.cast(tf.linalg.matrix_transpose(local_ct) <= tail_ct, local_mask.dtype)
            local_mask = tf.minimum(local_mask, ct_mask)

    local_k = _make_local_block(k, depth_k, batch, heads_kv, num_blocks,
                                block_length)
    local_v = _make_local_block(v, depth_v, batch, heads_kv, num_blocks,
                                block_length)
    tail_q = tf.slice(q, [0, 0, 1, 0, 0], [-1, -1, -1, -1, -1])
    tail_q = tf.reshape(tail_q,
                        [batch, heads, num_blocks - 1, block_length, depth_k])
    local_length = t2t_common.shape_list(local_k)[3]

    # make sure source_pos <= target_pos
    good_part = t2t_common.ones_matrix_band_part(
        block_length,
        local_length,
        -1,
        block_length,
        out_shape=[1, 1, 1, block_length, local_length])
    if mask is not None:
        bias = (1.0 - local_mask) * -1e9
    else:
        bias = (1.0 - good_part) * -1e9

    # TODO(noam): figure out how to show a summary for the remaining blocks.
    # The naive way currently causes errors due to empty tensors.
    # output: [batch, heads, num_blocks-1, block_length, depth_v]
    tail_output, tail_dict = dot_product_attention(
        tail_q,
        local_k,
        local_v,
        bias=bias,
        dropout_rate=dropout_rate,
        name="tail_block")
    tail_output = tf.reshape(
        tail_output, [batch, heads, (num_blocks - 1) * block_length, depth_v])
    output = tf.concat([first_output, tail_output], axis=2)

    # Remove the padding if introduced.
    output = tf.slice(output, [0, 0, 0, 0], [-1, -1, original_length, -1])
    output = tf.reshape(output, [batch, heads, original_length, depth_v])
    return output, {**first_dict, **tail_dict}


def combine_first_two_dimensions(x):
    """Reshape x so that the first two dimension become one.
    Args:
      x: a Tensor with shape [a, b, ...]
    Returns:
      a Tensor with shape [ab, ...]
    """
    ret = tf.reshape(x, tf.concat([[-1], t2t_common.shape_list(x)[2:]], 0))
    old_shape = x.get_shape().dims
    a, b = old_shape[:2]
    new_shape = [a * b if a and b else None] + old_shape[2:]
    ret.set_shape(new_shape)
    return ret


def dot_product_batched_head(q, k, v, gates_q, gates_k, mask_right=False):
    """Perform a dot product attention on a single sequence on a single head.
    This function dispatch the q, k, v and loop over the buckets to compute the
    attention dot product on each subsequences.
    Args:
      q (tf.Tensor): [batch*heads, length_q, depth_q]
      k (tf.Tensor): [batch*heads, length_kv, depth_q]
      v (tf.Tensor): [batch*heads, length_kv, depth_v]
      gates_q (tf.Tensor): One-hot of shape [batch*heads, length_q, nb_buckets]
      gates_k (tf.Tensor): One-hot of shape [batch*heads, length_k, nb_buckets]
      mask_right (bool): Add a bias to prevent attention to the future
    Returns:
      tf.Tensor: [length_q, depth_v]
    """
    nb_buckets = t2t_common.shape_list(gates_q)[-1]

    def get_dispatcher(gates):
        """Construct dispatcher for gates."""
        length = t2t_common.shape_list(gates)[1]
        # Count the number of ones per batch (and keep the max value)
        nb_elems_to_dispatch = tf.reduce_sum(gates, axis=[1, 2])
        nb_elems_to_dispatch = tf.reduce_max(nb_elems_to_dispatch)
        nb_elems_to_dispatch = to_int32(nb_elems_to_dispatch)
        capacity = nb_elems_to_dispatch // nb_buckets * 2  # Capacity is hardcoded
        capacity = tf.minimum(length, capacity)
        capacity = tf.maximum(capacity, 1)
        return t2t_expert_util.TruncatingDispatcher(gates, capacity)

    q_dispatcher = get_dispatcher(gates_q)
    k_dispatcher = get_dispatcher(gates_k)

    q = q_dispatcher.dispatch(q)
    k = k_dispatcher.dispatch(k)
    v = k_dispatcher.dispatch(v)

    # Bias of shape [batch*heads, nb_buckets, 1, capacity] broadcasted to every
    # queries
    bias = tf.expand_dims((k_dispatcher.nonpadding() - 1.0) * 1e9, 2)
    if mask_right:
        q_coordinate = t2t_common.to_float(
            tf.expand_dims(q_dispatcher.length_coordinate(), 3))
        k_coordinate = t2t_common.to_float(
            tf.expand_dims(k_dispatcher.length_coordinate(), 2))
        bias += t2t_common.to_float(tf.greater(k_coordinate, q_coordinate)) * -1e9
    # The sequence padding is not masked but is ignored on the next layers

    # q, k, v now have shape [batch*heads, nb_bucket, capacity, depth]
    # The buckets can be seen as different heads
    v_out, w_dict = dot_product_attention(q, k, v, bias=bias)

    # Combine all buckets together to restore the original length
    return q_dispatcher.combine(v_out), w_dict


def sparse_dot_product_attention_truncated(
        q,
        k,
        v,
        list_lsh,
        mask_right=False):  # pylint: disable=unused-argument
    """Sparse multihead self attention.
    Perform an approximation of the full multihead attention by dispatching
    the tokens using their keys/values. Thus the attention matrix are only
    computed each times on a subset of the tokens.
    Notes:
     * The function don't perform scaling here (multihead_attention does
    the /sqrt(depth)).
     * The padding should have been removed (so batch size should be 1 but length
     contains the elements from all different batches)
     * Right now, only self attention is supported so length_q and length_kv
     should be identical and the function will add triangular bias.
     * If bi.order is not None, The bias is added inside this function to
     prevent attention to the future.
    Args:
      q (tf.Tensor): Queries of shape [batch, heads, length_q, depth_k]
      k (tf.Tensor): Keys of shape [batch, heads, length_kv, depth_k]
      v (tf.Tensor): Values of shape [batch, heads, length_kv, depth_v]
      list_lsh: List of layers that can perform lsh-gating
      mask_right (bool):
    Returns:
      tf.Tensor: Approximation of Softmax(Q.K) * V, of shape
        [batch, heads, length_q, depth_v]
    """
    # Currently depth is the same for for q and v
    batch_size, nb_heads, _, depth = t2t_common.shape_list(q)
    _, nb_heads_kv, _, _ = t2t_common.shape_list(k)

    total_loss = 0.0

    # Each head get its own dispatcher
    # list_lsh = [LshGating(depth=depth, **experts_params) for _ in range(nb_heads)]

    def get_gates_head(x, add_first=False):
        """Return the gates for each heads of the current x.
        Args:
          x (tf.Tensor): of shape [batch, heads, length, depth]
          add_first (bool): if True, add the first element on each bucket
        Returns:
          tf.Tensor: gates of shape [batch, heads, length, num_buckets]
        """
        length = t2t_common.shape_list(x)[2]

        # Invert heads/batch
        x = tf.transpose(x, perm=[1, 0, 2, 3])
        x = tf.reshape(x, [nb_heads, batch_size * length, depth])

        list_x = tf.unstack(x)  # list[tf.Tensor(shape=[batch * length, depth])]

        # Unstack heads
        list_gates = []
        # There might be a more optimized way to compute all heads at once
        for lsh, single_x in zip(list_lsh, list_x):
            # Each head get its own dispatcher
            gates = lsh(single_x)
            nb_buckets = gates.get_shape().as_list()[-1]
            # Reshape to [batch, length, depth] but should consider sequence
            # padding in that case (also dispatch the padding)
            gates = tf.reshape(gates, [batch_size, length, nb_buckets])
            list_gates.append(gates)

        gates = tf.stack(list_gates)

        # Restore original shape
        gates = tf.reshape(gates, [nb_heads, batch_size, length, nb_buckets])
        gates = tf.transpose(gates, [1, 0, 2, 3])

        # Dispatch the first element to every gates to avoid empty buckets
        if add_first:
            gates = tf.maximum(gates,
                               tf.reshape(tf.one_hot([0], length), [1, 1, length, 1]))

        return gates

    gates_q = get_gates_head(q)
    gates_k = get_gates_head(k, add_first=True)

    # [batch, heads, length, depth] => [batch*heads, length, depth]
    q, k, v, gates_q, gates_k = [
        combine_first_two_dimensions(t) for t in (q, k, v, gates_q, gates_k)
    ]

    v_out, w_dict, = dot_product_batched_head(q, k, v, gates_q, gates_k, mask_right)

    # Restore original dimension
    v_out = tf.reshape(v_out, [batch_size, nb_heads, -1, depth])

    return v_out, total_loss / nb_heads, w_dict


def _matmul_with_relative_keys_2d(x, y, heads_share_relative_embedding):
    """Helper function for dot_product_unmasked_self_attention_relative_2d."""
    if heads_share_relative_embedding:
        ret = tf.einsum("bhxyd,md->bhxym", x, y)
    else:
        ret = tf.einsum("bhxyd,hmd->bhxym", x, y)
    return ret


def dot_product_unmasked_self_attention_relative_2d(
        q, k, v, bias,
        height_key_relative_embeddings,
        width_key_relative_embeddings,
        max_relative_position=None, dropout_rate=0.0, name=None,
        dropout_broadcast_dims=[0,1], heads_share_relative_embedding=False,
        add_relative_to_values=False):
    """Calculate relative position unmasked dot-product self-attention 2d.
    The attention calculation is augmented with learned representations for the
    relative position between each element in q and each element in k and v in
    height and width dimensions. for query index (i,j) and key index (l, m),
    the logit is q_i k_j^T + q_i rh_{l-i}^T + q_i rw_{m-j}^T, where rh and ry are
    the set of relative embeddings in height and width spatial dimensions,
    respectively.
    Args:
      q: a Tensor with shape [batch, heads, height, width, depth].
      k: a Tensor with shape [batch, heads, height, width, depth].
      v: a Tensor with shape [batch, heads, height, width, depth].
      bias: bias Tensor.
      max_relative_position: an integer the max relative embedding considered.
        Changing this invalidates checkpoints.
      dropout_rate: a floating point number.
      image_shapes: optional tuple of integer scalars.
      name: an optional string.
      make_image_summary: Whether to make an attention image summary.
      dropout_broadcast_dims:  an optional list of integers less than 4
        specifying in which dimensions to broadcast the dropout decisions.
        saves memory.
      heads_share_relative_embedding: a boolean indicating wheather to share
        relative embeddings between attention heads.
      add_relative_to_values: a boolean for adding relative embeddings to values.
    Returns:
      [batch, heads, height, width, depth] tensor, the output of attention.
      height_key_relative_embeddings: a 3d or 2d tensor, depending on head sharing
        settings, which are the relative embeddings for height.
      width_key_relative_embeddings: a 3d or 2d tensor, depending on head sharing
        settings, which are the relative embeddings for width.
    Raises:
      ValueError: if max_relative_position is not > 0.
    """
    if not max_relative_position:
        raise ValueError("Max relative position (%s) should be > 0 when using "
                         "relative self attention." % (max_relative_position))

    if add_relative_to_values:
        raise ValueError("Adding relative embeddings to values is not implemented")

    # This calculation only works for self attention.
    # q, k and v must therefore have the same shape.
    q.get_shape()[2:-1].assert_is_compatible_with(k.get_shape()[2:-1])
    q.get_shape()[2:-1].assert_is_compatible_with(v.get_shape()[2:-1])

    (height, width) = (t2t_common.shape_list(q)[2],
                       t2t_common.shape_list(q)[3])
    q_shape = t2t_common.shape_list(q)
    k_shape = t2t_common.shape_list(k)
    num_heads_q = q_shape[1]
    num_heads_k = k_shape[1]
    mqa = num_heads_q > num_heads_k
    depth_k = k_shape[-1]
    depth_v = t2t_common.shape_list(v)[-1]
    # flatten height width
    flatten_hw = lambda x, d, num_heads: tf.reshape(x, [-1, num_heads, height * width, d])
    # [batch, num_heads, query_length, memory_length]
    logits = tf.matmul(flatten_hw(q, depth_k, num_heads_q), flatten_hw(k, depth_k, num_heads_k),
                       transpose_b=True)

    def _compute_2d_relative_logits(
            query, key_relative_embeddings, height, width,
            heads_share_relative_embedding, transpose_mask, num_heads):
        """compute relative logits."""
        unmasked_rel_logits = _matmul_with_relative_keys_2d(
            query, key_relative_embeddings, heads_share_relative_embedding)
        # collapse height and heads
        unmasked_rel_logits = tf.reshape(unmasked_rel_logits,
                                         [-1, num_heads * height, width,
                                          2 * width - 1])
        unmasked_rel_logits = (
            _relative_position_to_absolute_position_unmasked(
                unmasked_rel_logits))
        # shape it back for tiling
        unmasked_rel_logits = tf.reshape(
            unmasked_rel_logits, [-1, num_heads, height, width, width])
        # tiling it height times
        unmasked_rel_logits = tf.expand_dims(
            unmasked_rel_logits, axis=3)
        unmasked_rel_logits = tf.tile(unmasked_rel_logits,
                                      [1, 1, 1, height, 1, 1])
        # bringing it to the right shape for adding to the logits.
        unmasked_rel_logits = tf.transpose(unmasked_rel_logits, transpose_mask)
        unmasked_rel_logits = tf.reshape(unmasked_rel_logits,
                                         [-1, num_heads, height * width,
                                          height * width])
        return unmasked_rel_logits

    # Relative logits in width dimension first.
    width_key_relative_embeddings = get_relative_embeddings_left_right(
        width_key_relative_embeddings,
        max_relative_position, width, depth_k,
        heads_share_relative_embedding)
    # [batch, heads, height, 2*width-1, 2*width-1]
    width_unmasked_rel_logits = _compute_2d_relative_logits(
        q, width_key_relative_embeddings, height, width,
        heads_share_relative_embedding, [0, 1, 2, 4, 3, 5], num_heads=num_heads_q)
    logits += width_unmasked_rel_logits
    # Relative logits in height dimension next. For ease, we transpose
    # height and width and repeat the above steps, and transpose to eventually
    # put the logits in their right positions.
    # [batch, heads, height, 2*height-1, 2*width-1]
    height_key_relative_embeddings = get_relative_embeddings_left_right(
        height_key_relative_embeddings,
        max_relative_position, height, depth_k,
        heads_share_relative_embedding)

    height_unmasked_rel_logits = _compute_2d_relative_logits(
        tf.transpose(q, [0, 1, 3, 2, 4]),
        height_key_relative_embeddings,
        width,
        height,
        heads_share_relative_embedding, [0, 1, 4, 2, 5, 3], num_heads=num_heads_q)
    logits += height_unmasked_rel_logits
    if bias is not None:
        logits += bias
    weights = tf.nn.softmax(logits, name="attention_weights")
    # dropping out the attention links for each of the heads
    weights = t2t_common.dropout_with_broadcast_dims(
        weights, dropout_rate, broadcast_dims=dropout_broadcast_dims)

    ret = tf.matmul(weights, flatten_hw(v, depth_v, num_heads_k))
    # reshape back the same spatial dimensions as q
    return tf.reshape(ret, [-1, num_heads_q, height, width, depth_v]), weights


def get_timing_signal_1d(length,
                         channels,
                         min_timescale=1.0,
                         max_timescale=1.0e4,
                         start_index=0):
    """Gets a bunch of sinusoids of different frequencies.
    Each channel of the input Tensor is incremented by a sinusoid of a different
    frequency and phase.
    This allows attention to learn to use absolute and relative positions.
    Timing signals should be added to some precursors of both the query and the
    memory inputs to attention.
    The use of relative position is possible because sin(x+y) and cos(x+y) can be
    expressed in terms of y, sin(x) and cos(x).
    In particular, we use a geometric sequence of timescales starting with
    min_timescale and ending with max_timescale.  The number of different
    timescales is equal to channels / 2. For each timescale, we
    generate the two sinusoidal signals sin(timestep/timescale) and
    cos(timestep/timescale).  All of these sinusoids are concatenated in
    the channels dimension.
    Args:
      length: scalar, length of timing signal sequence.
      channels: scalar, size of timing embeddings to create. The number of
          different timescales is equal to channels / 2.
      min_timescale: a float
      max_timescale: a float
      start_index: index of first position
    Returns:
      a Tensor of timing signals [1, length, channels]
    """
    position = t2t_common.to_float(tf.range(1, 1 + length) + start_index)
    num_timescales = channels // 2
    log_timescale_increment = (
            math.log(float(max_timescale) / float(min_timescale)) /
            tf.maximum(t2t_common.to_float(num_timescales) - 1, 1))
    inv_timescales = min_timescale * tf.exp(
        t2t_common.to_float(tf.range(num_timescales)) * -log_timescale_increment)
    scaled_time = tf.expand_dims(position, 1) * tf.expand_dims(inv_timescales, 0)
    # Please note that this slightly differs from the published paper.
    # See a discussion here: https://github.com/tensorflow/tensor2tensor/pull/177
    signal = tf.concat([tf.sin(scaled_time), tf.cos(scaled_time)], axis=1)
    signal = tf.pad(signal, [[0, 0], [0, tf.math.mod(channels, 2)]])
    signal = tf.reshape(signal, [1, length, channels])
    return signal


def add_timing_signal_1d(x,
                         min_timescale=1.0,
                         max_timescale=1.0e4,
                         start_index=0):
    """Adds a bunch of sinusoids of different frequencies to a Tensor.
    Each channel of the input Tensor is incremented by a sinusoid of a different
    frequency and phase.
    This allows attention to learn to use absolute and relative positions.
    Timing signals should be added to some precursors of both the query and the
    memory inputs to attention.
    The use of relative position is possible because sin(x+y) and cos(x+y) can be
    expressed in terms of y, sin(x) and cos(x).
    In particular, we use a geometric sequence of timescales starting with
    min_timescale and ending with max_timescale.  The number of different
    timescales is equal to channels / 2. For each timescale, we
    generate the two sinusoidal signals sin(timestep/timescale) and
    cos(timestep/timescale).  All of these sinusoids are concatenated in
    the channels dimension.
    Args:
      x: a Tensor with shape [batch, length, channels]
      min_timescale: a float
      max_timescale: a float
      start_index: index of first position
    Returns:
      a Tensor the same shape as x.
    """
    length = t2t_common.shape_list(x)[1]
    channels = t2t_common.shape_list(x)[2]
    signal = get_timing_signal_1d(length, channels, min_timescale, max_timescale,
                                  start_index)
    return x + t2t_common.cast_like(signal, x)



import abc

import numpy as np
import tensorflow as tf

tfal = None
try:
    import tensorflow_addons.layers as tfal
except:
    print('WARNING! TensorFlow Addons are missing!')
    tfal = None

from dynastes.layers.base_layers import DynastesBaseLayer
from dynastes.layers.conditioning_layers import FeaturewiseLinearModulation, ModulationLayer
from dynastes.ops.t2t_common import shape_list


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class PoolNormalization2D(DynastesBaseLayer):
    """
    My (Göran Sandström) own invention, performs smooth local response normalization
    using downsampling and upsampling
    """

    def __init__(self,
                 pool_size=(2, 2),
                 method=tf.image.ResizeMethod.BILINEAR, antialias=True,
                 **kwargs):
        super(PoolNormalization2D, self).__init__(**kwargs)
        self.pool_size = pool_size
        self.method = method
        self.antialias = antialias

    def call(self, x, training=None):
        orig_dtype = x.dtype
        x = tf.cast(x, tf.float32)
        x_size = shape_list(x)[1:-1]
        x_size = np.where(np.array(list(self.pool_size)) == -1, 1, x_size).tolist()
        x_size = tf.convert_to_tensor(x_size)
        reduce_axes = np.where(np.array(self.pool_size) == -1)[0].tolist()
        if len(reduce_axes) == 2:
            x -= tf.reduce_mean(x, axis=reduce_axes, keepdims=True)
            x *= tf.math.rsqrt(tf.reduce_mean(tf.square(x), axis=reduce_axes, keepdims=True) + 1e-8)
            x = tf.cast(x, orig_dtype)
            return x
        pool_size_t = tf.convert_to_tensor(self.pool_size)
        pool_size_t = tf.maximum(pool_size_t, 1)
        pooled_size = x_size // pool_size_t

        def pool_reduce(x, dtype=tf.float32):
            if len(reduce_axes) > 0:
                x = tf.reduce_mean(x, axis=reduce_axes, keepdims=True)
            x = tf.cast(tf.image.resize(tf.image.resize(x,
                                                        pooled_size,
                                                        method=self.method,
                                                        antialias=self.antialias),
                                        x_size,
                                        method=self.method,
                                        antialias=self.antialias), dtype)
            return x

        x -= pool_reduce(x, tf.float32)
        x *= tf.math.rsqrt(pool_reduce(tf.square(x), tf.float32) + 1e-8)
        x = tf.cast(x, orig_dtype)
        return x

    def get_config(self):
        config = {
            'pool_size': self.pool_size,
            'method': self.method,
            'antialias': self.antialias,
        }
        base_config = super(PoolNormalization2D, self).get_config()
        return {**base_config, **config}

    def compute_output_shape(self, input_shape):
        return input_shape


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class MultiNormalization(DynastesBaseLayer):
    def __init__(self,
                 layers,
                 **kwargs):
        super(MultiNormalization, self).__init__(**kwargs)
        self.norm_layers = layers

    def build(self, input_shape):
        for layer in self.norm_layers:
            layer.build(input_shape)
        self.add_weight('balance', [input_shape[-1], len(self.norm_layers)], trainable=True,
                        constraint=lambda x: tf.math.softmax(x, axis=-1))
        super(MultiNormalization, self).build(input_shape)

    def call(self, inputs, training=None):
        x = tf.stack([layer(inputs, training=training) for layer in self.norm_layers], axis=-1)
        x *= self.get_weight('balance', training=training)
        x = tf.reduce_sum(x, axis=-1)
        return x


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class ModulatedNormalization(DynastesBaseLayer, abc.ABC):

    def __init__(self, modulation_layer: ModulationLayer,
                 norm_layer, **kwargs):
        super(ModulatedNormalization, self).__init__(**kwargs)
        self.modulation_layer = modulation_layer
        self.norm_layer = norm_layer

    def build(self, input_shape):
        assert isinstance(input_shape, list)
        self.modulation_layer.build(input_shape)
        self.norm_layer.build(input_shape[0])
        super(ModulatedNormalization, self).build(input_shape)

    def call(self, inputs, training=None):
        orig_dtype = inputs[0].dtype
        normalized = self.norm_layer(inputs[0], training=training)
        return tf.cast(self.modulation_layer([normalized] + inputs[1:], training=training), orig_dtype)

    def compute_output_shape(self, input_shape):
        assert isinstance(input_shape, list)
        return input_shape[0]


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class AdaptiveNormalization(ModulatedNormalization, abc.ABC):

    def __init__(self, norm_layer, method=tf.image.ResizeMethod.BILINEAR, antialias=True, **kwargs):
        super(AdaptiveNormalization, self).__init__(
            modulation_layer=FeaturewiseLinearModulation(method=method, antialias=antialias),
            norm_layer=norm_layer,
            **kwargs)

    def get_config(self):
        config = {
            'method': self.method,
            'antialias': self.antialias,
            'mode': self.mode,
        }
        base_config = super(AdaptiveNormalization, self).get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class AdaptiveGroupNormalization(AdaptiveNormalization):

    def __init__(self, n_groups=2,
                 **kwargs):
        super(AdaptiveGroupNormalization, self).__init__(
            tfal.normalizations.GroupNormalization(groups=n_groups, center=False, scale=False), **kwargs)
        self.n_groups = n_groups

    def get_config(self):
        config = {
            'n_groups': self.n_groups,
        }
        base_config = super(AdaptiveGroupNormalization, self).get_config()
        return {**base_config, **config}


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class AdaptiveInstanceNormalization(AdaptiveGroupNormalization):
    """
    Introduced in:
    Arbitrary Style Transfer in Real-time with Adaptive Instance Normalization
    https://arxiv.org/abs/1703.06868

    Used in (notable):
    A Style-Based Generator Architecture for Generative Adversarial Networks ("StyleGAN")
    https://arxiv.org/abs/1812.04948

    Can also perform "Spatially-Adaptive Normalization":
    Semantic Image Synthesis with Spatially-Adaptive Normalization
    https://arxiv.org/abs/1903.07291
    """

    def __init__(self,
                 **kwargs):
        kwargs["groups"] = -1
        super(AdaptiveInstanceNormalization, self).__init__(**kwargs)


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class AdaptiveLayerNormalization(AdaptiveGroupNormalization):

    def __init__(self,
                 **kwargs):
        kwargs["groups"] = 1
        super(AdaptiveLayerNormalization, self).__init__(**kwargs)


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class AdaptiveMultiNormalization(AdaptiveNormalization):

    def __init__(self,
                 layers,
                 **kwargs):
        super(AdaptiveMultiNormalization, self).__init__(MultiNormalization(layers), **kwargs)


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class AdaptiveLayerInstanceNormalization(AdaptiveMultiNormalization):
    """
    Introduced in:
    U-GAT-IT: Unsupervised Generative Attentional Networks with Adaptive Layer-Instance Normalization for Image-to-Image Translation
    https://arxiv.org/abs/1907.10830
    """

    def __init__(self,
                 **kwargs):
        kwargs["layers"] = [tfal.normalizations.GroupNormalization(groups=1, center=False, scale=False),
                            tfal.normalizations.GroupNormalization(groups=-1, center=False, scale=False)]
        super(AdaptiveLayerInstanceNormalization, self).__init__(**kwargs)



def call_masked(layer, inputs, training=None, mask=None, **kwargs):
    try:
        return layer.call_masked(inputs, training=training, mask=mask, **kwargs)
    except AttributeError:
        try:
            out = layer(inputs, training=training, mask=mask, **kwargs)
        except Exception as e:
            out = layer(inputs, training=training, **kwargs)
        out_mask = compute_mask_if_possible(layer, inputs, mask=mask)
        return out, out_mask


def compute_mask_if_possible(layer, inputs, mask=None):
    if layer.supports_masking:
        out_mask = layer.compute_mask(inputs, mask)
    else:
        out_mask = mask
    return out_mask



import copy
from abc import ABCMeta
from functools import partial
from typing import List

import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp

tfpb = tfp.bijectors


def _get_roll_op(axis=0, steps=1, constant=0):
    def roll_op(x):
        return tf.stack([tf.roll(_x, -((i * steps) + constant), axis) for i, _x in enumerate(tf.unstack(x, axis=axis))],
                        axis=axis)

    return roll_op


class RollIncremental(tfpb.Bijector, metaclass=ABCMeta):

    def __init__(
            self,
            axis,
            steps=0,
            constant=0,
            **kwargs):
        kwargs['forward_min_event_ndims'] = axis + 1
        kwargs['inverse_min_event_ndims'] = axis + 1
        super(RollIncremental, self).__init__(**kwargs)
        setattr(self, '_forward', _get_roll_op(axis=axis, steps=steps, constant=constant))
        setattr(self, '_inverse', _get_roll_op(axis=axis, steps=-steps, constant=-constant))
        setattr(self, 'forward_event_shape', lambda x: x)
        setattr(self, 'inverse_event_shape', lambda x: x)


def _fix_unknown_dimension(input_shape, output_shape):
    """Find and replace a missing dimension in an output shape.
    This is a near direct port of the internal Numpy function
    `_fix_unknown_dimension` in `numpy/core/src/multiarray/shape.c`
    Arguments:
      input_shape: Shape of array being reshaped
      output_shape: Desired shape of the array with at most
        a single -1 which indicates a dimension that should be
        derived from the input shape.
    Returns:
      The new output shape with a -1 replaced with its computed value.
    Raises:
      ValueError: If the total array size of the output_shape is
      different than the input_shape, or more than one unknown dimension
      is specified.
    """
    output_shape = list(output_shape)
    msg = 'total size of new array must be unchanged'

    known, unknown = 1, None
    for index, dim in enumerate(output_shape):
        if dim < 0:
            if unknown is None:
                unknown = index
            else:
                raise ValueError('Can only specify one unknown dimension.')
        else:
            known *= dim

    original = np.prod(input_shape, dtype=int)
    if unknown is not None:
        if known == 0 or original % known != 0:
            raise ValueError(msg)
        output_shape[unknown] = original // known
    elif original != known:
        raise ValueError(msg)
    return output_shape


def _process_event_shapes(event_shape_in, event_shape_out, ch):
    event_shape_out = np.array(event_shape_out)
    event_shape_in = np.array(event_shape_in)
    infer_indices = np.where(event_shape_out == -1)[0]
    if len(infer_indices) == 0:
        return event_shape_out.tolist()
    elif len(infer_indices) == 1:
        if event_shape_out[-1] == -1:
            event_shape_out[-1] = ch
            return event_shape_out
        return _fix_unknown_dimension(event_shape_in.tolist(), event_shape_out.tolist())
    else:
        if event_shape_out[-1] == -1:
            event_shape_out[-1] = ch
        return _process_event_shapes(event_shape_in.tolist(), event_shape_out.tolist(), ch)


class EventShapeAwareChain(tfpb.Chain):

    def __init__(self,
                 event_shape_in,
                 partial_bijectors: List[partial],
                 **kwargs):

        bijectors = []
        ch = event_shape_in[-1]
        for partial_bijector in partial_bijectors:
            if partial_bijector.func == tfpb.Reshape:
                partial_bijector = partial(partial_bijector.func,
                                           **{k: copy.copy(v) for k, v in partial_bijector.keywords.items()})
                partial_bijector.keywords['event_shape_out'] = _process_event_shapes(event_shape_in,
                                                                                     partial_bijector.keywords[
                                                                                         'event_shape_out'], ch)
                bijector = partial_bijector(event_shape_in=event_shape_in)
            else:
                bijector = partial_bijector()
            event_shape_in = bijector.forward_event_shape(event_shape_in)
            bijectors.append(bijector)
        kwargs['bijectors'] = list(reversed(bijectors))
        super(EventShapeAwareChain, self).__init__(**kwargs)



import numpy as np
import six
import tensorflow as tf
import tensorflow.keras.layers as tfkl
from tensorflow.python.keras.utils.generic_utils import serialize_keras_object, deserialize_keras_object

from dynastes.ops.t2t_common import shape_list
from dynastes.weight_normalizers.spectral import SpectralNormalization


def serialize(normalizer):
    return serialize_keras_object(normalizer)


def deserialize(config, custom_objects={}):
    custom_objects = {**custom_objects, **object_scope}
    return deserialize_keras_object(
        config,
        module_objects=globals(),
        custom_objects=custom_objects,
        printable_module_name='normalizer')


def get(identifier):
    if identifier is None:
        return None
    if isinstance(identifier, dict):
        return deserialize(identifier)
    elif isinstance(identifier, six.string_types):
        identifier = str(identifier)
        # We have to special-case functions that return classes.
        # TODO(omalleyt): Turn these into classes or class aliases.
        if identifier == 'spectral':
            return SpectralNormalization()
    elif callable(identifier):
        return identifier
    else:
        raise ValueError('Could not interpret normalizer identifier:', identifier)


@tf.keras.utils.register_keras_serializable(package='Dynastes')
class WscaleNormalizer(tfkl.Layer):

    def __init__(self,
                 lrmul=1.,
                 gain=np.sqrt(2),
                 next_layer=tfkl.Activation('linear'),
                 trainable=False,
                 **kwargs):
        if 'trainable' in kwargs:
            kwargs.pop('trainable')
        super(WscaleNormalizer, self).__init__(trainable=False, **kwargs)
        self.next_layer = get(next_layer)
        self.gain = gain
        self.lrmul = lrmul

    def call(self, inputs, training=None):
        input_shape = shape_list(inputs)
        if len(input_shape) > 1:
            fan_in = np.prod(input_shape[:-1])  # [kernel, kernel, fmaps_in, fmaps_out] or [in, out]
            he_std = self.gain / np.sqrt(fan_in)  # He init
            runtime_coef = he_std * self.lrmul
        else:
            runtime_coef = self.lrmul
        return self.next_layer(inputs * runtime_coef, training=training)

    def get_config(self):
        config = {
            'lrmul': self.lrmul,
            'gain': self.gain,
            'next_layer': serialize(self.next_layer)
        }
        base_config = super(WscaleNormalizer, self).get_config()
        return {**base_config, **config}


object_scope = {
    'SpectralNormalization': SpectralNormalization,
    'WscaleNormalizer': WscaleNormalizer,
    'Activation': tfkl.Activation
}
